import uuid
from collections.abc import Iterable, Sequence
from typing import Self

import strawberry
from strawberry import ID, UNSET, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.artifact_registry_meta import ArtifactRegistryMetaConnection
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.data.artifact_registries.types import (
    ArtifactRegistryCreatorMeta,
    ArtifactRegistryModifierMeta,
)
from ai.backend.manager.data.reservoir_registry.types import ReservoirRegistryData
from ai.backend.manager.errors.api import NotImplementedAPI
from ai.backend.manager.models.reservoir_registry import ReservoirRegistryRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.reservoir_registry import ReservoirRegistryCreatorSpec
from ai.backend.manager.repositories.reservoir_registry.updaters import ReservoirRegistryUpdaterSpec
from ai.backend.manager.services.artifact_registry.actions.reservoir.create import (
    CreateReservoirRegistryAction,
)
from ai.backend.manager.services.artifact_registry.actions.reservoir.delete import (
    DeleteReservoirRegistryAction,
)
from ai.backend.manager.services.artifact_registry.actions.reservoir.get import (
    GetReservoirRegistryAction,
)
from ai.backend.manager.services.artifact_registry.actions.reservoir.get_multi import (
    GetReservoirRegistriesAction,
)
from ai.backend.manager.services.artifact_registry.actions.reservoir.list import (
    ListReservoirRegistriesAction,
)
from ai.backend.manager.services.artifact_registry.actions.reservoir.update import (
    UpdateReservoirRegistryAction,
)
from ai.backend.manager.types import OptionalState

from .types import StrawberryGQLContext


@strawberry.type(description="Added in 25.14.0")
class ReservoirRegistry(Node):
    id: NodeID[str]
    name: str
    endpoint: str
    access_key: str
    secret_key: str
    api_version: str

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.reservoir_registry_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ReservoirRegistryData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            endpoint=data.endpoint,
            access_key=data.access_key,
            secret_key=data.secret_key,
            api_version=data.api_version,
        )

    @classmethod
    async def load_by_id(
        cls, ctx: StrawberryGQLContext, reservoir_ids: Sequence[uuid.UUID]
    ) -> list["ReservoirRegistry"]:
        action_result = (
            await ctx.processors.artifact_registry.get_reservoir_registries.wait_for_complete(
                GetReservoirRegistriesAction(registry_ids=list(reservoir_ids))
            )
        )

        reservoirs = []
        for reservoir in action_result.result:
            reservoirs.append(ReservoirRegistry.from_dataclass(reservoir))

        return reservoirs

    @strawberry.field
    @classmethod
    async def remote_artifact_registries(
        cls, ctx: strawberry.Info[StrawberryGQLContext]
    ) -> ArtifactRegistryMetaConnection:
        raise NotImplementedAPI("This API is not implemented.")


ReservoirRegistryEdge = Edge[ReservoirRegistry]


@strawberry.type(description="Added in 25.14.0")
class ReservoirRegistryConnection(Connection[ReservoirRegistry]):
    @strawberry.field
    def count(self) -> int:
        return len(self.edges)


@strawberry.field(description="Added in 25.14.0")  # type: ignore[misc]
async def reservoir_registry(id: ID, info: Info[StrawberryGQLContext]) -> ReservoirRegistry | None:
    processors = info.context.processors
    action_result = await processors.artifact_registry.get_reservoir_registry.wait_for_complete(
        GetReservoirRegistryAction(reservoir_id=uuid.UUID(id))
    )
    return ReservoirRegistry.from_dataclass(action_result.result)


@strawberry.field(description="Added in 25.14.0")  # type: ignore[misc]
async def reservoir_registries(
    info: Info[StrawberryGQLContext],
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    offset: int | None = None,
    limit: int | None = None,
) -> ReservoirRegistryConnection | None:
    # TODO: Support pagination with before, after, first, last
    # TODO: Does we need to support filtering, ordering here?
    processors = info.context.processors

    action_result = await processors.artifact_registry.list_reservoir_registries.wait_for_complete(
        ListReservoirRegistriesAction()
    )

    nodes = [ReservoirRegistry.from_dataclass(data) for data in action_result.data]
    edges = [ReservoirRegistryEdge(node=node, cursor=encode_cursor(node.id)) for node in nodes]

    return ReservoirRegistryConnection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )


@strawberry.input(description="Added in 25.14.0")
class CreateReservoirRegistryInput:
    name: str
    endpoint: str
    access_key: str
    secret_key: str
    api_version: str

    def to_creator(self) -> Creator[ReservoirRegistryRow]:
        return Creator(
            spec=ReservoirRegistryCreatorSpec(
                endpoint=self.endpoint,
                access_key=self.access_key,
                secret_key=self.secret_key,
                api_version=self.api_version,
            )
        )

    def to_creator_meta(self) -> ArtifactRegistryCreatorMeta:
        return ArtifactRegistryCreatorMeta(name=self.name)


@strawberry.input(description="Added in 25.14.0")
class UpdateReservoirRegistryInput:
    id: ID
    name: str | None = UNSET
    endpoint: str | None = UNSET
    access_key: str | None = UNSET
    secret_key: str | None = UNSET
    api_version: str | None = UNSET

    def to_updater(self) -> Updater[ReservoirRegistryRow]:
        spec = ReservoirRegistryUpdaterSpec(
            endpoint=OptionalState[str].from_graphql(self.endpoint),
            access_key=OptionalState[str].from_graphql(self.access_key),
            secret_key=OptionalState[str].from_graphql(self.secret_key),
            api_version=OptionalState[str].from_graphql(self.api_version),
        )
        return Updater(spec=spec, pk_value=uuid.UUID(self.id))

    def to_modifier_meta(self) -> ArtifactRegistryModifierMeta:
        return ArtifactRegistryModifierMeta(
            name=OptionalState[str].from_graphql(self.name),
        )


@strawberry.input(description="Added in 25.14.0")
class DeleteReservoirRegistryInput:
    id: ID


@strawberry.type(description="Added in 25.14.0")
class CreateReservoirRegistryPayload:
    reservoir: ReservoirRegistry


@strawberry.type(description="Added in 25.14.0")
class UpdateReservoirRegistryPayload:
    reservoir: ReservoirRegistry


@strawberry.type(description="Added in 25.14.0")
class DeleteReservoirRegistryPayload:
    id: ID


@strawberry.mutation(description="Added in 25.14.0")  # type: ignore[misc]
async def create_reservoir_registry(
    input: CreateReservoirRegistryInput, info: Info[StrawberryGQLContext]
) -> CreateReservoirRegistryPayload:
    processors = info.context.processors

    action_result = await processors.artifact_registry.create_reservoir_registry.wait_for_complete(
        CreateReservoirRegistryAction(
            input.to_creator(),
            input.to_creator_meta(),
        )
    )

    return CreateReservoirRegistryPayload(
        reservoir=ReservoirRegistry.from_dataclass(action_result.result)
    )


@strawberry.mutation(description="Added in 25.14.0")  # type: ignore[misc]
async def update_reservoir_registry(
    input: UpdateReservoirRegistryInput, info: Info[StrawberryGQLContext]
) -> UpdateReservoirRegistryPayload:
    processors = info.context.processors

    action_result = await processors.artifact_registry.update_reservoir_registry.wait_for_complete(
        UpdateReservoirRegistryAction(updater=input.to_updater(), meta=input.to_modifier_meta())
    )

    return UpdateReservoirRegistryPayload(
        reservoir=ReservoirRegistry.from_dataclass(action_result.result)
    )


@strawberry.mutation(description="Added in 25.14.0")  # type: ignore[misc]
async def delete_reservoir_registry(
    input: DeleteReservoirRegistryInput, info: Info[StrawberryGQLContext]
) -> DeleteReservoirRegistryPayload:
    processors = info.context.processors

    await processors.artifact_registry.delete_reservoir_registry.wait_for_complete(
        DeleteReservoirRegistryAction(reservoir_id=uuid.UUID(input.id))
    )

    return DeleteReservoirRegistryPayload(id=input.id)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact.types import (
    ArtifactDataWithRevisions,
    ArtifactFilterOptions,
    ArtifactOrderingOptions,
)
from ai.backend.manager.services.artifact.actions.base import ArtifactAction
from ai.backend.manager.types import PaginationOptions


@dataclass
class ListArtifactsWithRevisionsAction(ArtifactAction):
    """Action to list artifacts with revisions using old-style pagination (REST API)."""

    pagination: PaginationOptions | None = None
    ordering: ArtifactOrderingOptions | None = None
    filters: ArtifactFilterOptions | None = None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class ListArtifactsWithRevisionsActionResult(BaseActionResult):
    data: list[ArtifactDataWithRevisions]
    # Note: Total number of artifacts, this is not equals to len(data)
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None



"""GraphQL types for model replicas."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from typing import Any, Self, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo

from ai.backend.common.data.model_deployment.types import ActivenessStatus as CommonActivenessStatus
from ai.backend.common.data.model_deployment.types import LivenessStatus as CommonLivenessStatus
from ai.backend.common.data.model_deployment.types import ReadinessStatus as CommonReadinessStatus
from ai.backend.manager.api.gql.base import (
    OrderDirection,
    to_global_id,
)
from ai.backend.manager.api.gql.session import Session
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.api.gql_legacy.session import ComputeSessionNode
from ai.backend.manager.data.deployment.types import (
    ModelReplicaData,
    ReplicaOrderField,
    RouteStatus,
    RouteTrafficStatus,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.deployment.options import RouteConditions, RouteOrders
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_id import (
    GetRevisionByIdAction,
)

from .revision import ModelRevision

# ========== Enums ==========

ReadinessStatus: type[CommonReadinessStatus] = strawberry.enum(
    CommonReadinessStatus,
    name="ReadinessStatus",
    description="Added in 25.19.0. This enum represents the readiness status of a replica, indicating whether the deployment has been checked and its health state.",
)

LivenessStatus: type[CommonLivenessStatus] = strawberry.enum(
    CommonLivenessStatus,
    name="LivenessStatus",
    description="Added in 25.19.0. This enum represents the liveness status of a replica, indicating whether the deployment is currently running and able to serve requests.",
)

ActivenessStatus: type[CommonActivenessStatus] = strawberry.enum(
    CommonActivenessStatus,
    name="ActivenessStatus",
    description="Added in 25.19.0. This enum represents the activeness status of a replica, indicating whether the deployment is currently active and able to serve requests.",
)

ReplicaStatus: type[RouteStatus] = strawberry.enum(
    RouteStatus,
    name="ReplicaStatus",
    description="Added in 25.19.0. This enum represents the provisioning status of a replica.",
)

TrafficStatus: type[RouteTrafficStatus] = strawberry.enum(
    RouteTrafficStatus,
    name="TrafficStatus",
    description="Added in 25.19.0. This enum represents the traffic status of a replica.",
)


# ========== Status Filters ==========


@strawberry.input(description="Added in 25.19.0")
class ReplicaStatusFilter:
    in_: list[ReplicaStatus] | None = strawberry.field(name="in", default=None)
    equals: ReplicaStatus | None = None


@strawberry.input(description="Added in 25.19.0")
class TrafficStatusFilter:
    in_: list[TrafficStatus] | None = strawberry.field(name="in", default=None)
    equals: TrafficStatus | None = None


# ========== ModelReplica Types ==========


@strawberry.input(description="Added in 25.19.0")
class ReplicaFilter(GQLFilter):
    status: ReplicaStatusFilter | None = None
    traffic_status: TrafficStatusFilter | None = None

    AND: list[ReplicaFilter] | None = None
    OR: list[ReplicaFilter] | None = None
    NOT: list[ReplicaFilter] | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.status:
            if self.status.in_ is not None:
                statuses = [RouteStatus(s) for s in self.status.in_]
                conditions.append(RouteConditions.by_statuses(statuses))
            elif self.status.equals is not None:
                conditions.append(RouteConditions.by_statuses([RouteStatus(self.status.equals)]))

        if self.traffic_status:
            if self.traffic_status.in_ is not None:
                traffic_statuses = [RouteTrafficStatus(s) for s in self.traffic_status.in_]
                conditions.append(RouteConditions.by_traffic_statuses(traffic_statuses))
            elif self.traffic_status.equals is not None:
                conditions.append(
                    RouteConditions.by_traffic_statuses([
                        RouteTrafficStatus(self.traffic_status.equals)
                    ])
                )

        return conditions


@strawberry.input(description="Added in 25.19.0")
class ReplicaOrderBy(GQLOrderBy):
    field: ReplicaOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ReplicaOrderField.CREATED_AT:
                return RouteOrders.created_at(ascending)
            case ReplicaOrderField.ID:
                return RouteOrders.id(ascending)


@strawberry.type
class ModelReplica(Node):
    """
    Added in 25.19.0.

    Represents a single replica instance of a model deployment. Each replica
    runs in a separate compute session and serves inference requests.

    Replicas have health status indicators (readiness, liveness, activeness)
    and traffic weight for load balancing.
    """

    id: NodeID[str]
    _session_id: strawberry.Private[UUID]
    _revision_id: strawberry.Private[UUID]
    readiness_status: ReadinessStatus = strawberry.field(
        description="Whether the replica has been checked and its health state.",
    )
    liveness_status: LivenessStatus = strawberry.field(
        description="Whether the replica is currently running and able to serve requests.",
    )
    activeness_status: ActivenessStatus = strawberry.field(
        description="Whether the replica is currently active and able to serve requests.",
    )
    weight: int = strawberry.field(
        description="Traffic weight for load balancing between replicas."
    )
    created_at: datetime = strawberry.field(description="Timestamp when the replica was created.")

    @strawberry.field(  # type: ignore[misc]
        description="The session ID associated with the replica. This can be null right after replica creation."
    )
    async def session(self, info: Info[StrawberryGQLContext]) -> Session:
        session_global_id = to_global_id(
            ComputeSessionNode, self._session_id, is_target_graphene_object=True
        )
        return Session(id=ID(session_global_id))

    @strawberry.field
    async def revision(self, info: Info[StrawberryGQLContext]) -> ModelRevision:
        """Resolve revision by ID."""
        processor = info.context.processors.deployment
        result = await processor.get_revision_by_id.wait_for_complete(
            GetRevisionByIdAction(revision_id=self._revision_id)
        )
        return ModelRevision.from_dataclass(result.data)

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.replica_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ModelReplicaData) -> Self:
        return cls(
            id=ID(str(data.id)),
            _revision_id=data.revision_id,
            _session_id=data.session_id,
            readiness_status=ReadinessStatus(data.readiness_status),
            liveness_status=LivenessStatus(data.liveness_status),
            activeness_status=ActivenessStatus(data.activeness_status),
            weight=data.weight,
            created_at=data.created_at,
        )


ModelReplicaEdge = Edge[ModelReplica]


@strawberry.type
class ModelReplicaConnection(Connection[ModelReplica]):
    """
    Added in 25.19.0.

    A Relay-style connection for paginated access to model replicas.
    Includes total count for UI pagination display.
    """

    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count

    @classmethod
    def from_dataclass(cls, replicas_data: list[ModelReplicaData]) -> ModelReplicaConnection:
        nodes = [ModelReplica.from_dataclass(data) for data in replicas_data]
        edges = [ModelReplicaEdge(node=node, cursor=str(node.id)) for node in nodes]

        page_info = PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        )

        return cls(count=len(nodes), edges=edges, page_info=page_info)


@strawberry.type(description="Added in 25.19.0")
class ReplicaStatusChangedPayload:
    replica: ModelReplica



from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, override

from ai.backend.manager.data.error_log.types import ErrorLogSeverity
from ai.backend.manager.models.error_logs import ErrorLogRow
from ai.backend.manager.repositories.base import CreatorSpec

__all__ = ("ErrorLogCreatorSpec",)


@dataclass
class ErrorLogCreatorSpec(CreatorSpec[ErrorLogRow]):
    severity: ErrorLogSeverity
    source: str
    message: str
    context_lang: str
    context_env: dict[str, Any]
    user: uuid.UUID | None = None
    is_read: bool = False
    is_cleared: bool = False
    request_url: str | None = None
    request_status: int | None = None
    traceback: str | None = None
    created_at: datetime | None = None

    @override
    def build_row(self) -> ErrorLogRow:
        return ErrorLogRow(
            severity=self.severity,
            source=self.source,
            message=self.message,
            context_lang=self.context_lang,
            context_env=self.context_env,
            user=self.user,
            is_read=self.is_read,
            is_cleared=self.is_cleared,
            request_url=self.request_url,
            request_status=self.request_status,
            traceback=self.traceback,
            created_at=self.created_at,
        )



from collections.abc import Callable
from enum import Enum
from typing import Any, NamedTuple, TypeVar

import sqlalchemy as sa


class ArrayFieldItem(NamedTuple):
    column_name: str


class JSONFieldItem(NamedTuple):
    column_name: str
    key_name: str


class ORMFieldItem(NamedTuple):
    column: sa.orm.attributes.InstrumentedAttribute[Any] | sa.Column[Any]


TEnum = TypeVar("TEnum", bound=Enum)


class EnumFieldItem[TEnum: Enum](NamedTuple):
    column_name: str
    enum_cls: TEnum


FieldSpecItem = tuple[
    str | ArrayFieldItem | JSONFieldItem | EnumFieldItem[Any] | ORMFieldItem,
    Callable[[str], Any] | None,
]
OrderSpecItem = tuple[
    str | ArrayFieldItem | JSONFieldItem | EnumFieldItem[Any],
    Callable[[sa.Column[Any]], Any] | None,
]


def get_col_from_table(
    table: sa.Table | sa.sql.Join | type, column_name: str
) -> (
    sa.Column[Any]
    | sa.orm.attributes.InstrumentedAttribute[Any]
    | sa.sql.elements.KeyedColumnElement[Any]
):
    if isinstance(table, (sa.Table, sa.sql.Join)):
        col: sa.Column[Any] | sa.sql.elements.KeyedColumnElement[Any] = table.c[column_name]
        return col
    # For ORM class table
    attr: sa.orm.attributes.InstrumentedAttribute[Any] = getattr(table, column_name)
    return attr


class ExternalTableFilterSpec:
    """
    Specification for filtering on external tables that require JOINs.
    This allows filters on related tables (like project_name from GroupRow)
    to be handled separately and passed to repository layer for JOIN operations.
    """

    def __init__(
        self,
        field_name: str,
        target_table: sa.Table,
        target_column: str,
        join_builder: Callable[[sa.Table | sa.sql.Join], sa.sql.Join],
        transform: Callable[[str], Any] | None = None,
    ) -> None:
        """
        Args:
            field_name: Name of the field in the filter expression (e.g., "project_name")
            target_table: SQLAlchemy table to apply the filter on (e.g., GroupRow.__table__)
            target_column: Column name in the target table (e.g., "name")
            join_builder: Function that builds the JOIN clause given the base table or existing join
            transform: Optional transform function for the field value
        """
        self.field_name = field_name
        self.target_table = target_table
        self.target_column = target_column
        self.transform = transform
        self.join_builder = join_builder



from __future__ import annotations

import uuid
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column

from ai.backend.manager.models.base import (
    GUID,
    Base,
)


class EntityFieldRow(Base):  # type: ignore[misc]
    __tablename__ = "entity_fields"
    __table_args__ = (
        sa.UniqueConstraint(
            "entity_type",
            "entity_id",
            "field_type",
            "field_id",
            name="uq_entity_fields_mapping",
        ),
        sa.Index("ix_entity_fields_entity_lookup", "entity_type", "entity_id"),
        sa.Index("ix_entity_fields_field_lookup", "field_type", "field_id"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )

    entity_type: Mapped[str] = mapped_column(
        "entity_type",
        sa.String(64),
        nullable=False,
    )
    entity_id: Mapped[str] = mapped_column(
        "entity_id",
        sa.String(64),
        nullable=False,
    )
    field_type: Mapped[str] = mapped_column(
        "field_type",
        sa.String(64),
        nullable=False,
    )
    field_id: Mapped[str] = mapped_column(
        "field_id",
        sa.String(64),
        nullable=False,
    )
    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )



# This module contains legacy Graphene-based GraphQL schema.
# It is being replaced by the Strawberry-based implementation in api/gql.
#
# NOTE: We intentionally avoid star imports here to prevent circular imports
# when models import utility functions from gql_legacy.base.



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from datetime import datetime

from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.types import (
    AgentSelectionStrategy,
    ResourceSlot,
    SessionId,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.metrics.scheduler import (
    SchedulerPhaseMetricObserver,
)
from ai.backend.manager.repositories.fair_share import FairShareRepository
from ai.backend.manager.repositories.scheduler import (
    SchedulerRepository,
    SchedulingData,
)
from ai.backend.manager.sokovan.data import (
    AllocationBatch,
    KeypairOccupancy,
    SchedulingFailure,
    SchedulingPredicate,
    SessionAllocation,
    SessionWorkload,
    SystemSnapshot,
)
from ai.backend.manager.sokovan.recorder import (
    ExecutionRecord,
    RecorderContext,
    StepStatus,
)
from ai.backend.manager.sokovan.scheduler.results import ScheduleResult

from .allocators.allocator import SchedulingAllocator
from .selectors.concentrated import ConcentratedAgentSelector
from .selectors.dispersed import DispersedAgentSelector
from .selectors.legacy import LegacyAgentSelector
from .selectors.roundrobin import RoundRobinAgentSelector
from .selectors.selector import (
    AgentInfo,
    AgentSelectionConfig,
    AgentSelectionCriteria,
    AgentSelector,
    KernelResourceSpec,
    SessionMetadata,
)
from .sequencers.drf import DRFSequencer
from .sequencers.fair_share import FairShareSequencer
from .sequencers.fifo import FIFOSequencer
from .sequencers.lifo import LIFOSequencer
from .sequencers.sequencer import SchedulingSequencer, WorkloadSequencer
from .validators.validator import SchedulingValidator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class SessionProvisionerArgs:
    validator: SchedulingValidator
    default_sequencer: WorkloadSequencer
    default_agent_selector: AgentSelector
    allocator: SchedulingAllocator
    repository: SchedulerRepository
    fair_share_repository: FairShareRepository
    config_provider: ManagerConfigProvider
    valkey_schedule: ValkeyScheduleClient


class SessionProvisioner:
    """
    Handles the PENDING -> SCHEDULED transition for sessions.

    Orchestrates the provisioning pipeline:
    1. Validation (via validators)
    2. Sequencing (via sequencers)
    3. Agent selection (via selectors)
    4. Resource allocation (via allocators)
    """

    _validator: SchedulingValidator
    _default_sequencer: WorkloadSequencer
    _default_agent_selector: AgentSelector
    _allocator: SchedulingAllocator
    _repository: SchedulerRepository
    _fair_share_repository: FairShareRepository
    _config_provider: ManagerConfigProvider
    _sequencer_pool: Mapping[str, WorkloadSequencer]
    _agent_selector_pool: Mapping[AgentSelectionStrategy, AgentSelector]
    _phase_metrics: SchedulerPhaseMetricObserver
    _valkey_schedule: ValkeyScheduleClient

    def __init__(self, args: SessionProvisionerArgs) -> None:
        self._validator = args.validator
        self._default_sequencer = args.default_sequencer
        self._default_agent_selector = args.default_agent_selector
        self._allocator = args.allocator
        self._repository = args.repository
        self._fair_share_repository = args.fair_share_repository
        self._config_provider = args.config_provider
        self._valkey_schedule = args.valkey_schedule
        self._sequencer_pool = self._make_sequencer_pool()
        self._agent_selector_pool = self._make_agent_selector_pool(
            args.config_provider.config.manager.agent_selection_resource_priority
        )
        self._phase_metrics = SchedulerPhaseMetricObserver.instance()

    def _make_sequencer_pool(self) -> Mapping[str, WorkloadSequencer]:
        """Initialize the sequencer pool with default sequencers."""
        pool: dict[str, WorkloadSequencer] = defaultdict(DRFSequencer)
        pool["fifo"] = FIFOSequencer()
        pool["lifo"] = LIFOSequencer()
        pool["drf"] = DRFSequencer()
        pool["fair-share"] = FairShareSequencer(self._fair_share_repository)
        return pool

    @classmethod
    def _make_agent_selector_pool(
        cls, agent_selection_resource_priority: list[str]
    ) -> Mapping[AgentSelectionStrategy, AgentSelector]:
        """Initialize the agent selector pool with default selectors."""
        pool: dict[AgentSelectionStrategy, AgentSelector] = defaultdict(
            lambda: AgentSelector(ConcentratedAgentSelector(agent_selection_resource_priority))
        )
        pool[AgentSelectionStrategy.CONCENTRATED] = AgentSelector(
            ConcentratedAgentSelector(agent_selection_resource_priority)
        )
        pool[AgentSelectionStrategy.DISPERSED] = AgentSelector(
            DispersedAgentSelector(agent_selection_resource_priority)
        )
        pool[AgentSelectionStrategy.ROUNDROBIN] = AgentSelector(RoundRobinAgentSelector())
        pool[AgentSelectionStrategy.LEGACY] = AgentSelector(
            LegacyAgentSelector(agent_selection_resource_priority)
        )
        return pool

    def _get_sequencer(self, name: str) -> SchedulingSequencer:
        sequncer = self._sequencer_pool[name]
        return SchedulingSequencer(sequncer)

    async def schedule_scaling_group(
        self,
        scaling_group: str,
        scheduling_data: SchedulingData,
        provision_time: datetime,
    ) -> ScheduleResult:
        """
        Schedule sessions for a specific scaling group.

        This method orchestrates the full provisioning pipeline using pre-fetched data:
        1. Sequencing: Order workloads using configured sequencer (FIFO/LIFO/DRF)
        2. Validation: Check quotas and constraints for each workload
        3. Agent selection: Select agents using configured strategy
        4. Allocation: Persist allocations to database

        Args:
            scaling_group: The scaling group to schedule for.
            scheduling_data: Pre-fetched scheduling data from Handler.

        Returns:
            ScheduleResult containing scheduled session data
        """
        # Use data from scheduling_data instead of making DB calls
        # Convert PendingSessionData to SessionWorkload
        workloads = [
            session.to_session_workload() for session in scheduling_data.pending_sessions.sessions
        ]
        sg_info = scheduling_data.scaling_group

        if not scheduling_data.snapshot_data:
            log.warning("Missing snapshot data for scaling group {}", scaling_group)
            return ScheduleResult()

        # Convert snapshot data to SystemSnapshot
        system_snapshot = scheduling_data.snapshot_data.to_system_snapshot(
            scheduling_data.spec.known_slot_types, scheduling_data.total_capacity
        )

        # Create agent selection config directly from spec and scaling group opts
        selection_config = AgentSelectionConfig(
            max_container_count=scheduling_data.spec.max_container_count,
            enforce_spreading_endpoint_replica=sg_info.scheduler_opts.enforce_spreading_endpoint_replica,
        )
        # Perform sequencing (batch operation for all workloads)
        # Record as shared phase so all entity records include it
        sequencer = self._get_sequencer(sg_info.scheduler)
        with (
            self._phase_metrics.measure_phase(
                "scheduler", scaling_group, f"sequencing_{sg_info.scheduler}"
            ),
            RecorderContext[SessionId].shared_phase(
                "sequencing", success_detail=sequencer.success_message()
            ),
            RecorderContext[SessionId].shared_step(
                sequencer.name, success_detail=sequencer.success_message()
            ),
        ):
            sequenced_workloads = await sequencer.sequence(
                scaling_group, system_snapshot, workloads
            )

        # Build mutable agents with occupancy data from snapshot
        agent_occupancy = (
            scheduling_data.snapshot_data.resource_occupancy.by_agent
            if scheduling_data.snapshot_data
            else {}
        )
        mutable_agents = [
            AgentInfo.from_meta_and_occupancy(agent, agent_occupancy)
            for agent in scheduling_data.agents
        ]
        session_allocations: list[SessionAllocation] = []
        scheduling_failures: list[SchedulingFailure] = []
        # Get agent selection strategy from scheduler opts config
        agent_selection_strategy = sg_info.scheduler_opts.agent_selection_strategy
        agent_selector = self._agent_selector_pool[agent_selection_strategy]

        # Get current pool from RecorderContext (scope opened by coordinator)
        pool = RecorderContext[SessionId].current_pool()

        for session_workload in sequenced_workloads:
            try:
                # Sequencing phase is automatically included via shared phases
                session_allocation = await self._schedule_workload(
                    scaling_group,
                    system_snapshot,
                    mutable_agents,
                    selection_config,
                    agent_selector,
                    session_workload,
                )
                session_allocations.append(session_allocation)
            except Exception as e:
                log.debug(
                    "Scheduling failed for workload {}: {}",
                    session_workload.session_id,
                    e,
                )
                # Get execution record from pool and convert to SchedulingFailure
                record = pool.get_record(session_workload.session_id)
                passed, failed = self._convert_record_to_predicates(record)
                failure = SchedulingFailure(
                    session_id=session_workload.session_id,
                    passed_phases=passed,
                    failed_phases=failed,
                    last_try=provision_time,
                    msg=str(e),
                )
                scheduling_failures.append(failure)
                continue

        # Convert execution records to passed_phases/failed_phases for allocations
        for allocation in session_allocations:
            record = pool.get_record(allocation.session_id)
            if record:
                passed, failed = self._convert_record_to_predicates(record)
                allocation.passed_phases = passed
                allocation.failed_phases = failed

        log.info(
            "Processing {} allocations and {} failures in scaling group {}",
            len(session_allocations),
            len(scheduling_failures),
            scaling_group,
        )
        # Create batch with allocations and failures
        batch = AllocationBatch(
            allocations=session_allocations,
            failures=scheduling_failures,
        )
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "allocation"):
            scheduled_sessions = await self._allocator.allocate(batch)

        failure_ids = [f.session_id for f in scheduling_failures]
        await self._valkey_schedule.set_pending_queue(scaling_group, failure_ids)
        return ScheduleResult(
            scheduled_sessions=scheduled_sessions,
        )

    async def _schedule_workload(
        self,
        scaling_group: str,
        mutable_snapshot: SystemSnapshot,
        mutable_agents: Sequence[AgentInfo],
        selection_config: AgentSelectionConfig,
        agent_selector: AgentSelector,
        session_workload: SessionWorkload,
    ) -> SessionAllocation:
        pool = RecorderContext[SessionId].current_pool()
        recorder = pool.recorder(session_workload.session_id)

        # Phase 1: Validation
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "validation"):
            with recorder.phase("validation"):
                self._validator.validate(mutable_snapshot, session_workload)

        # Phase 2: Agent Selection
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "agent_selection"):
            with recorder.phase(
                "agent_selection", success_detail=agent_selector.strategy_success_message()
            ):
                with recorder.step(
                    agent_selector.strategy_name(),
                    success_detail=agent_selector.strategy_success_message(),
                ):
                    session_allocation = await self._allocate_workload(
                        session_workload,
                        mutable_agents,
                        selection_config,
                        scaling_group,
                        agent_selector,
                    )

        # Phase 3: Allocation (prepare)
        with recorder.phase("allocation", success_detail=self._allocator.success_message()):
            with recorder.step(
                self._allocator.name(), success_detail=self._allocator.success_message()
            ):
                # Update the snapshot to reflect this allocation
                # Note: agent state changes are already applied to mutable_agents
                self._update_system_snapshot(
                    mutable_snapshot,
                    session_workload,
                    session_allocation,
                )

        return session_allocation

    def _update_system_snapshot(
        self,
        snapshot: SystemSnapshot,
        workload: SessionWorkload,
        allocation: SessionAllocation,
    ) -> None:
        """
        Update the system snapshot after a session allocation.
        This ensures the next validation uses up-to-date information.

        :param snapshot: The system snapshot to update (modified in-place)
        :param workload: The session workload that was allocated
        :param allocation: The session allocation result containing agent allocations
        """
        # Calculate total allocated resources from allocation
        total_allocated_slots = ResourceSlot()
        for agent_alloc in allocation.agent_allocations:
            for slot in agent_alloc.allocated_slots:
                total_allocated_slots += slot

        # 1. Update resource occupancy - add the session's allocated slots
        # Update keypair occupancy
        current_keypair = snapshot.resource_occupancy.by_keypair.get(workload.access_key)
        if current_keypair is None:
            current_keypair = KeypairOccupancy(
                occupied_slots=ResourceSlot(), session_count=0, sftp_session_count=0
            )

        # Update occupied slots and session counts
        current_keypair.occupied_slots += total_allocated_slots
        if workload.is_private:
            current_keypair.sftp_session_count += 1
        else:
            current_keypair.session_count += 1

        snapshot.resource_occupancy.by_keypair[workload.access_key] = current_keypair

        # Update user occupancy
        current_user = snapshot.resource_occupancy.by_user.get(workload.user_uuid, ResourceSlot())
        snapshot.resource_occupancy.by_user[workload.user_uuid] = (
            current_user + total_allocated_slots
        )

        # Update group occupancy
        current_group = snapshot.resource_occupancy.by_group.get(workload.group_id, ResourceSlot())
        snapshot.resource_occupancy.by_group[workload.group_id] = (
            current_group + total_allocated_slots
        )

        # Update domain occupancy
        current_domain = snapshot.resource_occupancy.by_domain.get(
            workload.domain_name, ResourceSlot()
        )
        snapshot.resource_occupancy.by_domain[workload.domain_name] = (
            current_domain + total_allocated_slots
        )

        # 2. Update concurrency counts
        if workload.is_private:
            # Increment SFTP session count
            current_sftp = snapshot.concurrency.sftp_sessions_by_keypair.get(workload.access_key, 0)
            snapshot.concurrency.sftp_sessions_by_keypair[workload.access_key] = current_sftp + 1
        else:
            # Increment regular session count
            current_sessions = snapshot.concurrency.sessions_by_keypair.get(workload.access_key, 0)
            snapshot.concurrency.sessions_by_keypair[workload.access_key] = current_sessions + 1

    async def _allocate_workload(
        self,
        session_workload: SessionWorkload,
        agents_info: Sequence[AgentInfo],
        selection_config: AgentSelectionConfig,
        scaling_group: str,
        agent_selector: AgentSelector,
    ) -> SessionAllocation:
        """
        Allocate resources for a single session workload.

        :param session_workload: The workload to allocate
        :param agents_info: Available agents (will be modified with updated states)
        :param selection_config: Agent selection configuration
        :param scaling_group: The scaling group name
        :return: SessionAllocation
        :raises AgentSelectionError: If agent selection fails
        """
        # Convert session workload to agent selection criteria
        session_metadata = SessionMetadata(
            session_id=session_workload.session_id,
            session_type=session_workload.session_type,
            scaling_group=session_workload.scaling_group,
            cluster_mode=session_workload.cluster_mode,
        )

        kernel_requirements = {
            kernel.kernel_id: KernelResourceSpec(
                requested_slots=kernel.requested_slots,
                required_architecture=kernel.architecture,
            )
            for kernel in session_workload.kernels
        }

        criteria = AgentSelectionCriteria(
            session_metadata=session_metadata,
            kernel_requirements=kernel_requirements,
            kernel_counts_at_endpoint=session_workload.kernel_counts_at_endpoint,
        )

        # Use batch selection method - it will get resource requirements internally
        # and apply state changes to agents_info
        selections = await agent_selector.select_agents_for_batch_requirements(
            agents_info,
            criteria,
            selection_config,
            session_workload.designated_agent_ids,
        )

        # Build session allocation from selections
        return SessionAllocation.from_agent_selections(
            session_workload,
            selections,
            scaling_group,
        )

    @staticmethod
    def _convert_record_to_predicates(
        record: ExecutionRecord | None,
    ) -> tuple[list[SchedulingPredicate], list[SchedulingPredicate]]:
        """
        Convert an ExecutionRecord to passed/failed SchedulingPredicate lists.

        Args:
            record: The execution record to convert (may be None if entity context failed early)

        Returns:
            Tuple of (passed_predicates, failed_predicates)
        """
        passed: list[SchedulingPredicate] = []
        failed: list[SchedulingPredicate] = []

        if record is None:
            return passed, failed

        for phase in record.phases:
            # Add phase-level predicate
            phase_predicate = SchedulingPredicate(
                name=phase.name,
                msg=phase.detail or "",
            )
            if phase.status == StepStatus.SUCCESS:
                passed.append(phase_predicate)
            else:
                failed.append(phase_predicate)

            # Add step-level predicates
            for step in phase.steps:
                step_predicate = SchedulingPredicate(
                    name=step.name,
                    msg=step.detail or "",
                )
                if step.status == StepStatus.SUCCESS:
                    passed.append(step_predicate)
                else:
                    failed.append(step_predicate)

        return passed, failed



from __future__ import annotations

import strawberry
from strawberry import ID, Info


@strawberry.federation.type(keys=["id"], name="ComputeSessionNode", extend=True)
class Session:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID, info: Info) -> Session:
        return cls(id=id)



"""Revision resolver functions."""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.scalars import JSON

from ai.backend.manager.api.gql.base import resolve_global_id
from ai.backend.manager.api.gql.deployment.fetcher.revision import fetch_revisions
from ai.backend.manager.api.gql.deployment.types.deployment import ModelDeployment
from ai.backend.manager.api.gql.deployment.types.revision import (
    ActivateRevisionInputGQL,
    ActivateRevisionPayloadGQL,
    AddRevisionInput,
    AddRevisionPayload,
    CreateRevisionInput,
    CreateRevisionPayload,
    ModelRevision,
    ModelRevisionConnection,
    ModelRevisionFilter,
    ModelRevisionOrderBy,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.data.deployment.inference_runtime_config import (
    MOJORuntimeConfig,
    NVDIANIMRuntimeConfig,
    SGLangRuntimeConfig,
    VLLMRuntimeConfig,
)
from ai.backend.manager.services.deployment.actions.model_revision.add_model_revision import (
    AddModelRevisionAction,
)
from ai.backend.manager.services.deployment.actions.model_revision.create_model_revision import (
    CreateModelRevisionAction,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_id import (
    GetRevisionByIdAction,
)
from ai.backend.manager.services.deployment.actions.revision_operations.activate_revision import (
    ActivateRevisionAction,
)

# Query resolvers


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def revisions(
    info: Info[StrawberryGQLContext],
    filter: ModelRevisionFilter | None = None,
    order_by: list[ModelRevisionOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ModelRevisionConnection | None:
    """List revisions with optional filtering and pagination."""
    return await fetch_revisions(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def revision(id: ID, info: Info[StrawberryGQLContext]) -> ModelRevision | None:
    """Get a specific revision by ID."""
    _, revision_id = resolve_global_id(id)
    processor = info.context.processors.deployment
    result = await processor.get_revision_by_id.wait_for_complete(
        GetRevisionByIdAction(revision_id=UUID(revision_id))
    )
    return ModelRevision.from_dataclass(result.data)


@strawberry.field(  # type: ignore[misc]
    description="Added in 25.16.0. Get JSON Schema for inference runtime configuration"
)
async def inference_runtime_config(name: str) -> JSON:
    match name.lower():
        case "vllm":
            return VLLMRuntimeConfig.to_json_schema()
        case "sglang":
            return SGLangRuntimeConfig.to_json_schema()
        case "nvdianim":
            return NVDIANIMRuntimeConfig.to_json_schema()
        case "mojo":
            return MOJORuntimeConfig.to_json_schema()
        case _:
            return {
                "error": "Unknown service name",
            }


@strawberry.field(  # type: ignore[misc]
    description="Added in 25.16.0 Get configuration JSON Schemas for all inference runtimes"
)
async def inference_runtime_configs(info: Info[StrawberryGQLContext]) -> JSON:
    return {
        "vllm": VLLMRuntimeConfig.to_json_schema(),
        "sglang": SGLangRuntimeConfig.to_json_schema(),
        "nvdianim": NVDIANIMRuntimeConfig.to_json_schema(),
        "mojo": MOJORuntimeConfig.to_json_schema(),
    }


# Mutation resolvers


@strawberry.mutation(description="Added in 25.16.0")  # type: ignore[misc]
async def add_model_revision(
    input: AddRevisionInput, info: Info[StrawberryGQLContext]
) -> AddRevisionPayload:
    """Add a model revision to a deployment."""
    processor = info.context.processors.deployment
    result = await processor.add_model_revision.wait_for_complete(
        AddModelRevisionAction(
            model_deployment_id=UUID(input.deployment_id), adder=input.to_model_revision_creator()
        )
    )

    return AddRevisionPayload(revision=ModelRevision.from_dataclass(result.revision))


@strawberry.mutation(  # type: ignore[misc]
    description="Added in 25.16.0. Create model revision which is not attached to any deployment."
)
async def create_model_revision(
    input: CreateRevisionInput, info: Info[StrawberryGQLContext]
) -> CreateRevisionPayload:
    """Create a new model revision without attaching it to any deployment."""
    processor = info.context.processors.deployment
    result = await processor.create_model_revision.wait_for_complete(
        CreateModelRevisionAction(creator=input.to_model_revision_creator())
    )

    return CreateRevisionPayload(revision=ModelRevision.from_dataclass(result.revision))


@strawberry.mutation(  # type: ignore[misc]
    description="Added in 25.19.0. Activate a specific revision to be the current revision."
)
async def activate_deployment_revision(
    input: ActivateRevisionInputGQL,
    info: Info[StrawberryGQLContext, None],
) -> ActivateRevisionPayloadGQL:
    """Activate a revision to be the current revision for a deployment."""
    _, deployment_id = resolve_global_id(input.deployment_id)
    _, revision_id = resolve_global_id(input.revision_id)

    processor = info.context.processors.deployment
    result = await processor.activate_revision.wait_for_complete(
        ActivateRevisionAction(
            deployment_id=UUID(deployment_id),
            revision_id=UUID(revision_id),
        )
    )

    return ActivateRevisionPayloadGQL(
        deployment=ModelDeployment.from_dataclass(result.deployment),
        previous_revision_id=(
            ID(str(result.previous_revision_id)) if result.previous_revision_id else None
        ),
        activated_revision_id=ID(str(result.activated_revision_id)),
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .actions import CreateErrorLogAction, CreateErrorLogActionResult
from .actions.search import SearchErrorLogsAction, SearchErrorLogsActionResult

if TYPE_CHECKING:
    from ai.backend.manager.repositories.error_log import ErrorLogRepository

__all__ = ("ErrorLogService",)


@dataclass
class ErrorLogService:
    """Service for error log operations."""

    _repository: ErrorLogRepository

    def __init__(self, repository: ErrorLogRepository) -> None:
        self._repository = repository

    async def create(self, action: CreateErrorLogAction) -> CreateErrorLogActionResult:
        """Creates a new error log."""
        error_log_data = await self._repository.create(action.creator)
        return CreateErrorLogActionResult(error_log_data=error_log_data)

    async def search(self, action: SearchErrorLogsAction) -> SearchErrorLogsActionResult:
        """Search error logs with querier pattern."""
        result = await self._repository.search(action.querier)
        return SearchErrorLogsActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )



from __future__ import annotations

import uuid
from collections.abc import Collection
from typing import cast

import sqlalchemy as sa

from ai.backend.manager.models.error_logs import ErrorLogRow
from ai.backend.manager.repositories.base import QueryCondition


class ErrorLogConditions:
    """Query conditions for error logs."""

    @staticmethod
    def by_ids(error_log_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(sa.sql.expression.ColumnElement[bool], ErrorLogRow.id.in_(error_log_ids))

        return inner



from .row import (
    ADMIN_PERMISSIONS,
    ALL_DOMAIN_PERMISSIONS,
    MAXIMUM_DOTFILE_SIZE,
    MEMBER_PERMISSIONS,
    MONITOR_PERMISSIONS,
    OWNER_PERMISSIONS,
    PRIVILEGED_MEMBER_PERMISSIONS,
    DomainDotfile,
    DomainModel,
    DomainPermissionContext,
    DomainPermissionContextBuilder,
    DomainRow,
    domains,
    get_domains,
    get_permission_ctx,
    query_domain_dotfiles,
    row_to_data,
    verify_dotfile_name,
)

__all__ = (
    "ADMIN_PERMISSIONS",
    "ALL_DOMAIN_PERMISSIONS",
    "MAXIMUM_DOTFILE_SIZE",
    "MEMBER_PERMISSIONS",
    "MONITOR_PERMISSIONS",
    "OWNER_PERMISSIONS",
    "PRIVILEGED_MEMBER_PERMISSIONS",
    "DomainDotfile",
    "DomainModel",
    "DomainPermissionContext",
    "DomainPermissionContextBuilder",
    "DomainRow",
    "domains",
    "get_domains",
    "get_permission_ctx",
    "query_domain_dotfiles",
    "row_to_data",
    "verify_dotfile_name",
)



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.model_serving.types import EndpointData
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class ModifyEndpointAction(ModelServiceAction):
    endpoint_id: uuid.UUID
    updater: Updater[EndpointRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ModifyEndpointActionResult(BaseActionResult):
    success: bool
    data: EndpointData | None

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data is not None else None



from __future__ import annotations

import uuid
from collections.abc import Iterable, Mapping
from datetime import datetime
from typing import Any, Self

import strawberry
from aiotools import apartial
from strawberry import ID, UNSET, Info
from strawberry.dataloader import DataLoader
from strawberry.relay import Connection, Edge, Node, NodeID
from strawberry.scalars import JSON

from ai.backend.common.data.artifact.types import (
    VerificationStepResult,
    VerifierResult,
)
from ai.backend.common.data.storage.registries.types import ModelTarget as ModelTargetData
from ai.backend.manager.api.gql.artifact_registry_meta import ArtifactRegistryMeta
from ai.backend.manager.api.gql.base import (
    ByteSize,
    IntFilter,
    OrderDirection,
    StringFilter,
    UUIDFilter,
)
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.artifact.types import (
    ArtifactAvailability,
    ArtifactData,
    ArtifactOrderField,
    ArtifactRemoteStatus,
    ArtifactRevisionData,
    ArtifactRevisionOrderField,
    ArtifactStatus,
    ArtifactType,
)
from ai.backend.manager.data.artifact.types import DelegateeTarget as DelegateeTargetData
from ai.backend.manager.defs import ARTIFACT_MAX_SCAN_LIMIT
from ai.backend.manager.repositories.artifact.options import (
    ArtifactConditions,
    ArtifactOrders,
    ArtifactRevisionConditions,
    ArtifactRevisionOrders,
)
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.services.artifact.actions.get import GetArtifactAction
from ai.backend.manager.services.artifact_revision.actions.get import GetArtifactRevisionAction


@strawberry.type(
    name="ArtifactVerifierMetadataEntry",
    description=dedent_strip("""
    Added in 26.1.0.

    A single key-value entry representing metadata from an artifact verifier.
    Contains additional information about the verification process.
    """),
)
class ArtifactVerifierMetadataEntryGQL:
    """Single metadata entry with key and value."""

    key: str = strawberry.field(description="The key identifier for this metadata entry.")
    value: str = strawberry.field(description="The value for this metadata entry.")


@strawberry.type(
    name="ArtifactVerifierMetadata",
    description=dedent_strip("""
    Added in 26.1.0.

    A collection of metadata from an artifact verifier.
    Contains key-value pairs providing additional information about the verification.
    """),
)
class ArtifactVerifierMetadataGQL:
    """Metadata containing multiple key-value entries."""

    entries: list[ArtifactVerifierMetadataEntryGQL] = strawberry.field(
        description="List of metadata entries. Each entry contains a key-value pair."
    )

    @classmethod
    def from_mapping(cls, data: Mapping[str, str]) -> ArtifactVerifierMetadataGQL:
        """Convert a Mapping to GraphQL type."""
        entries = [ArtifactVerifierMetadataEntryGQL(key=k, value=v) for k, v in data.items()]
        return cls(entries=entries)


@strawberry.type(
    name="ArtifactVerifierResult",
    description=dedent_strip("""
    Added in 25.17.0.

    Result from a single malware verifier containing scan results and metadata.

    Each verifier scans the artifact for potential security issues and reports
    findings including infected file count, scan time, and any errors encountered.
    """),
)
class ArtifactVerifierGQLResult:
    success: bool = strawberry.field(description="Whether the verification completed successfully")
    infected_count: int = strawberry.field(
        description="Number of infected or suspicious files detected"
    )
    scanned_at: datetime = strawberry.field(description="Timestamp when verification started")
    scan_time: float = strawberry.field(
        description="Time taken to complete verification in seconds"
    )
    scanned_count: int = strawberry.field(description="Total number of files scanned")
    metadata: ArtifactVerifierMetadataGQL = strawberry.field(
        description="Added in 26.1.0. Additional metadata from the verifier."
    )
    error: str | None = strawberry.field(
        description="Fatal error message if the verifier failed to complete"
    )

    @classmethod
    def from_dataclass(cls, data: VerifierResult) -> Self:
        return cls(
            success=data.success,
            infected_count=data.infected_count,
            scanned_at=data.scanned_at,
            scan_time=data.scan_time,
            scanned_count=data.scanned_count,
            metadata=ArtifactVerifierMetadataGQL.from_mapping(data.metadata),
            error=data.error,
        )


@strawberry.type(
    name="ArtifactVerifierResultEntry",
    description=dedent_strip("""
    Added in 25.17.0.

    Entry for a single verifier's result in the verification results.

    Associates a verifier name with its scan results.
    """),
)
class ArtifactVerifierGQLResultEntry:
    name: str = strawberry.field(description="Name of the verifier (e.g., 'clamav', 'custom')")
    result: ArtifactVerifierGQLResult = strawberry.field(
        description="Scan result from this verifier"
    )

    @classmethod
    def from_verifier_result(cls, name: str, result: VerifierResult) -> Self:
        return cls(
            name=name,
            result=ArtifactVerifierGQLResult.from_dataclass(result),
        )


@strawberry.type(
    name="ArtifactVerificationResult",
    description=dedent_strip("""
    Added in 25.17.0.

    Complete verification result containing results from all configured verifiers.

    Artifacts undergo malware scanning through multiple verifiers after being imported.
    This type aggregates results from all verifiers that were run on the artifact.
    """),
)
class ArtifactVerificationGQLResult:
    verifiers: list[ArtifactVerifierGQLResultEntry] = strawberry.field(
        description="Results from each verifier that scanned the artifact"
    )

    @classmethod
    def from_dataclass(cls, data: VerificationStepResult) -> Self:
        verifier_entries = [
            ArtifactVerifierGQLResultEntry.from_verifier_result(verifier_name, verifier_result)
            for verifier_name, verifier_result in data.verifiers.items()
        ]
        return cls(verifiers=verifier_entries)


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Filter options for artifacts based on various criteria such as type, name, registry,
    source, and availability status.

    Supports logical operations (AND, OR, NOT) for complex filtering scenarios.
    """)
)
class ArtifactFilter(GQLFilter):
    type: list[ArtifactType] | None = None
    name: StringFilter | None = None
    registry: StringFilter | None = None
    source: StringFilter | None = None
    availability: list[ArtifactAvailability] | None = None

    AND: list[ArtifactFilter] | None = None
    OR: list[ArtifactFilter] | None = None
    NOT: list[ArtifactFilter] | None = None

    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list containing QueryConditions that represent
        all filters with proper logical operators applied.
        """
        # Collect direct field conditions (these will be combined with AND)
        field_conditions: list[QueryCondition] = []

        # Apply type filter
        if self.type:
            field_conditions.append(ArtifactConditions.by_types(self.type))

        # Apply name filter
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=ArtifactConditions.by_name_contains,
                equals_factory=ArtifactConditions.by_name_equals,
                starts_with_factory=ArtifactConditions.by_name_starts_with,
                ends_with_factory=ArtifactConditions.by_name_ends_with,
            )
            if name_condition:
                field_conditions.append(name_condition)

        # Apply registry filter
        if self.registry:
            registry_condition = self.registry.build_query_condition(
                contains_factory=ArtifactConditions.by_registry_contains,
                equals_factory=ArtifactConditions.by_registry_equals,
                starts_with_factory=ArtifactConditions.by_registry_starts_with,
                ends_with_factory=ArtifactConditions.by_registry_ends_with,
            )
            if registry_condition:
                field_conditions.append(registry_condition)

        # Apply source filter
        if self.source:
            source_condition = self.source.build_query_condition(
                contains_factory=ArtifactConditions.by_source_contains,
                equals_factory=ArtifactConditions.by_source_equals,
                starts_with_factory=ArtifactConditions.by_source_starts_with,
                ends_with_factory=ArtifactConditions.by_source_ends_with,
            )
            if source_condition:
                field_conditions.append(source_condition)

        # Apply availability filter
        if self.availability:
            field_conditions.append(ArtifactConditions.by_availability(self.availability))

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Specifies the field and direction for ordering artifacts in queries.
    """)
)
class ArtifactOrderBy(GQLOrderBy):
    field: ArtifactOrderField
    direction: OrderDirection = OrderDirection.ASC

    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ArtifactOrderField.NAME:
                return ArtifactOrders.name(ascending)
            case ArtifactOrderField.TYPE:
                return ArtifactOrders.type(ascending)
            case ArtifactOrderField.SCANNED_AT:
                return ArtifactOrders.scanned_at(ascending)
            case ArtifactOrderField.UPDATED_AT:
                return ArtifactOrders.updated_at(ascending)
            case ArtifactOrderField.SIZE:
                # SIZE ordering is not supported yet, fall back to updated_at
                return ArtifactOrders.updated_at(ascending)


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Filter for artifact revision status. Supports exact match or inclusion in a list of statuses.
    """)
)
class ArtifactRevisionStatusFilter:
    in_: list[ArtifactStatus] | None = strawberry.field(name="in", default=None)
    equals: ArtifactStatus | None = None


@strawberry.input(description="Added in 25.16.0")
class ArtifactRevisionRemoteStatusFilter:
    in_: list[ArtifactRemoteStatus] | None = strawberry.field(name="in", default=None)
    equals: ArtifactRemoteStatus | None = None


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Filter options for artifact revisions based on status, version, artifact ID, and file size.

    Supports logical operations (AND, OR, NOT) for complex filtering scenarios.
    """)
)
class ArtifactRevisionFilter(GQLFilter):
    status: ArtifactRevisionStatusFilter | None = None
    remote_status: ArtifactRevisionRemoteStatusFilter | None = strawberry.field(
        default=None, description="Added in 25.16.0"
    )
    version: StringFilter | None = None
    artifact_id: UUIDFilter | None = strawberry.field(default=None)
    size: IntFilter | None = None

    AND: list[ArtifactRevisionFilter] | None = None
    OR: list[ArtifactRevisionFilter] | None = None
    NOT: list[ArtifactRevisionFilter] | None = None

    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list containing QueryConditions that represent
        all filters with proper logical operators applied.
        """
        # Collect direct field conditions (these will be combined with AND)
        field_conditions: list[QueryCondition] = []

        # Apply status filter
        if self.status:
            statuses_to_filter: list[ArtifactStatus] = []
            if self.status.in_:
                statuses_to_filter = self.status.in_
            elif self.status.equals:
                statuses_to_filter = [self.status.equals]

            if statuses_to_filter:
                field_conditions.append(ArtifactRevisionConditions.by_statuses(statuses_to_filter))

        # Apply remote_status filter
        if self.remote_status:
            remote_statuses_to_filter: list[ArtifactRemoteStatus] = []
            if self.remote_status.in_:
                remote_statuses_to_filter = self.remote_status.in_
            elif self.remote_status.equals:
                remote_statuses_to_filter = [self.remote_status.equals]

            if remote_statuses_to_filter:
                field_conditions.append(
                    ArtifactRevisionConditions.by_remote_statuses(remote_statuses_to_filter)
                )

        # Apply version filter
        if self.version:
            version_condition = self.version.build_query_condition(
                contains_factory=ArtifactRevisionConditions.by_version_contains,
                equals_factory=ArtifactRevisionConditions.by_version_equals,
                starts_with_factory=ArtifactRevisionConditions.by_version_starts_with,
                ends_with_factory=ArtifactRevisionConditions.by_version_ends_with,
            )
            if version_condition:
                field_conditions.append(version_condition)

        # Apply artifact_id filter
        if self.artifact_id:
            if self.artifact_id.equals:
                field_conditions.append(
                    ArtifactRevisionConditions.by_artifact_id(self.artifact_id.equals)
                )
            elif self.artifact_id.in_:
                field_conditions.append(
                    ArtifactRevisionConditions.by_artifact_ids(self.artifact_id.in_)
                )

        # Apply size filter
        if self.size:
            if self.size.equals is not None:
                field_conditions.append(ArtifactRevisionConditions.by_size_equals(self.size.equals))
            elif self.size.not_equals is not None:
                field_conditions.append(
                    ArtifactRevisionConditions.by_size_not_equals(self.size.not_equals)
                )
            elif self.size.greater_than is not None:
                field_conditions.append(
                    ArtifactRevisionConditions.by_size_greater_than(self.size.greater_than)
                )
            elif self.size.greater_than_or_equal is not None:
                field_conditions.append(
                    ArtifactRevisionConditions.by_size_greater_than_or_equal(
                        self.size.greater_than_or_equal
                    )
                )
            elif self.size.less_than is not None:
                field_conditions.append(
                    ArtifactRevisionConditions.by_size_less_than(self.size.less_than)
                )
            elif self.size.less_than_or_equal is not None:
                field_conditions.append(
                    ArtifactRevisionConditions.by_size_less_than_or_equal(
                        self.size.less_than_or_equal
                    )
                )

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Specifies the field and direction for ordering artifact revisions in queries.
    """)
)
class ArtifactRevisionOrderBy(GQLOrderBy):
    field: ArtifactRevisionOrderField
    direction: OrderDirection = OrderDirection.ASC

    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ArtifactRevisionOrderField.VERSION:
                return ArtifactRevisionOrders.version(ascending)
            case ArtifactRevisionOrderField.STATUS:
                return ArtifactRevisionOrders.status(ascending)
            case ArtifactRevisionOrderField.SIZE:
                return ArtifactRevisionOrders.size(ascending)
            case ArtifactRevisionOrderField.CREATED_AT:
                return ArtifactRevisionOrders.created_at(ascending)
            case ArtifactRevisionOrderField.UPDATED_AT:
                return ArtifactRevisionOrders.updated_at(ascending)


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for scanning artifacts from external registries (HuggingFace, Reservoir).

    Discovers available artifacts and registers their metadata in the system.
    Artifacts remain in SCANNED status until explicitly imported via import_artifacts.
    """)
)
class ScanArtifactsInput:
    registry_id: ID | None = None
    limit: int = strawberry.field(
        description=f"Maximum number of artifacts to scan (max: {ARTIFACT_MAX_SCAN_LIMIT})"
    )
    artifact_type: ArtifactType | None = None
    search: str | None = None


@strawberry.input(
    description=dedent_strip("""
    Added in 26.1.0.

    Options for importing artifact revisions.

    Controls import behavior such as forcing re-download regardless of digest freshness.
    """)
)
class ImportArtifactsOptionsGQL:
    force: bool = strawberry.field(
        default=False,
        description="Force re-download regardless of digest freshness check.",
    )


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for importing scanned artifact revisions from external registries.

    Downloads artifact files from the external source and transitions them through:
    SCANNED → PULLING → PULLED → AVAILABLE status progression.
    """)
)
class ImportArtifactsInput:
    artifact_revision_ids: list[ID]
    vfolder_id: ID | None = strawberry.field(
        default=None,
        description="Target vfolder ID to store the imported artifacts. Added in 26.1.0.",
    )
    options: ImportArtifactsOptionsGQL | None = strawberry.field(
        default=None,
        description="Options controlling import behavior such as forcing re-download. Added in 26.1.0.",
    )


@strawberry.input(description="Added in 25.15.0")
class DelegateeTarget:
    delegatee_reservoir_id: ID
    target_registry_id: ID

    def to_dataclass(self) -> DelegateeTargetData:
        return DelegateeTargetData(
            delegatee_reservoir_id=uuid.UUID(self.delegatee_reservoir_id),
            target_registry_id=uuid.UUID(self.target_registry_id),
        )


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input type for delegated scanning of artifacts from a delegatee reservoir registry's remote registry.
""")
)
class DelegateScanArtifactsInput:
    delegator_reservoir_id: ID | None = strawberry.field(
        default=None, description="ID of the reservoir registry to delegate the scan request to"
    )
    delegatee_target: DelegateeTarget | None = strawberry.field(
        default=None,
        description="Target delegatee reservoir registry and its remote registry to scan",
    )
    limit: int = strawberry.field(
        description=f"Maximum number of artifacts to scan (max: {ARTIFACT_MAX_SCAN_LIMIT})"
    )
    artifact_type: ArtifactType | None = strawberry.field(
        default=None, description="Filter artifacts by type (e.g., model, image, package)"
    )
    search: str | None = strawberry.field(
        default=None, description="Search term to filter artifacts by name or description"
    )


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input type for delegated import of artifact revisions from a reservoir registry's remote registry.
    Used to specify which artifact revisions should be imported from the remote registry source
    into the local reservoir registry storage.
""")
)
class DelegateImportArtifactsInput:
    artifact_revision_ids: list[ID] = strawberry.field(
        description="List of artifact revision IDs of delegatee artifact registry"
    )
    delegator_reservoir_id: ID | None = strawberry.field(
        default=None, description="ID of the reservoir registry to delegate the import request to"
    )
    artifact_type: ArtifactType | None = strawberry.field(default=None)
    delegatee_target: DelegateeTarget | None = strawberry.field(default=None)
    options: ImportArtifactsOptionsGQL | None = strawberry.field(
        default=None,
        description="Options controlling import behavior such as forcing re-download. Added in 26.1.0.",
    )


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for updating artifact metadata properties.

    Modifies artifact metadata such as readonly status and description.
    This operation does not affect the actual artifact files or revisions.
    """)
)
class UpdateArtifactInput:
    artifact_id: ID
    readonly: bool | None = UNSET
    description: str | None = UNSET


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for canceling an in-progress artifact import operation.

    Stops the download process and reverts the artifact revision status back to SCANNED.
    """)
)
class CancelArtifactInput:
    artifact_revision_id: ID


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for cleaning up stored artifact revision data.

    Removes downloaded files from storage and transitions the artifact revision
    back to SCANNED status, freeing up storage space.
    """)
)
class CleanupArtifactRevisionsInput:
    artifact_revision_ids: list[ID]


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input for soft-deleting artifacts from the system.

    Marks artifacts as deleted without permanently removing them.
    Deleted artifacts can be restored using restore_artifacts.
    """)
)
class DeleteArtifactsInput:
    artifact_ids: list[ID]


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input for restoring previously deleted artifacts.

    Reverses the soft-delete operation, making the artifacts available again.
    """)
)
class RestoreArtifactsInput:
    artifact_ids: list[ID]


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for approving an artifact revision.

    Admin-only operation to approve artifact revisions for general use.
    """)
)
class ApproveArtifactInput:
    artifact_revision_id: ID


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for rejecting an artifact revision.

    Admin-only operation to reject artifact revisions, preventing their use.
    """)
)
class RejectArtifactInput:
    artifact_revision_id: ID


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for subscribing to artifact status change notifications.

    Used with artifact_status_changed subscription to receive real-time updates
    when artifact revision statuses change.
    """)
)
class ArtifactStatusChangedInput:
    artifact_revision_ids: list[ID]


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Specifies a target model for scanning operations.

    Used to identify specific models in external registries for detailed scanning.
    If revision is not specified, defaults to 'main' revision.
    """)
)
class ModelTarget:
    model_id: str
    revision: str | None = None

    def to_dataclass(self) -> ModelTargetData:
        return ModelTargetData(model_id=self.model_id, revision=self.revision)


@strawberry.input(
    description=dedent_strip("""
    Added in 25.14.0.

    Input for batch scanning of specific models from external registries.

    Scans multiple specified models and retrieves detailed information including
    README content and file sizes. This operation performs immediate detailed scanning
    unlike the general scan_artifacts which only retrieves basic metadata.
    """)
)
class ScanArtifactModelsInput:
    models: list[ModelTarget]
    registry_id: uuid.UUID | None = None


# Object Types
@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Information about the source or registry of an artifact.

    Contains the name and URL of the registry where the artifact is stored or originates from.
    """)
)
class SourceInfo:
    name: str | None
    url: str | None


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    An artifact represents AI models, packages, images, or other resources that can be
    stored, managed, and used within Backend.AI.

    Artifacts are discovered through scanning external registries and,
    can have multiple revisions.

    Each artifact contains metadata and references to its source registry.

    Key concepts:
    - Type: MODEL, PACKAGE, or IMAGE
    - Availability: ALIVE (available), DELETED (soft-deleted)
    - Source: Original external registry where it was discovered
    """)
)
class Artifact(Node):
    id: NodeID[str]
    name: str
    type: ArtifactType
    description: str | None
    registry: SourceInfo
    source: SourceInfo
    readonly: bool
    extra: JSON | None
    scanned_at: datetime
    updated_at: datetime
    availability: ArtifactAvailability

    @classmethod
    def from_dataclass(cls, data: ArtifactData, registry_url: str, source_url: str) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            type=ArtifactType(data.type),
            description=data.description,
            registry=SourceInfo(name=data.registry_type.value, url=registry_url),
            source=SourceInfo(name=data.source_registry_type.value, url=source_url),
            readonly=data.readonly,
            extra=data.extra,
            scanned_at=data.scanned_at,
            updated_at=data.updated_at,
            availability=data.availability,
        )

    @strawberry.field
    async def revisions(
        self,
        info: Info[StrawberryGQLContext],
        filter: ArtifactRevisionFilter | None = None,
        order_by: list[ArtifactRevisionOrderBy] | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> ArtifactRevisionConnection:
        from .fetcher import fetch_artifact_revisions

        base_conditions = [ArtifactRevisionConditions.by_artifact_id(uuid.UUID(self.id))]

        return await fetch_artifact_revisions(
            info,
            filter=filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
            base_conditions=base_conditions,
        )


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    A specific version/revision of an artifact containing the actual file data.

    Artifact revisions progress through different statuses:
    - SCANNED: Discovered in external registry, metadata only
    - PULLING: Currently downloading from external source
    - PULLED: Downloaded to temporary storage
    - AVAILABLE: Ready for use by users

    Contains version information, file size, README content, and timestamps.
    Most HuggingFace models only have a 'main' revision.
    """)
)
class ArtifactRevision(Node):
    id: NodeID[str]
    status: ArtifactStatus
    remote_status: ArtifactRemoteStatus | None = strawberry.field(description="Added in 25.15.0")
    version: str
    readme: str | None
    size: ByteSize | None
    created_at: datetime | None
    updated_at: datetime | None
    digest: str | None = strawberry.field(
        description="Digest at the time of import. None for models that have not been imported. Added in 25.17.0"
    )
    verification_result: ArtifactVerificationGQLResult | None = strawberry.field(
        description="Verification result containing malware scan results from all verifiers. None if not yet verified. Added in 25.17.0"
    )

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.artifact_revision_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ArtifactRevisionData) -> Self:
        return cls(
            id=ID(str(data.id)),
            status=ArtifactStatus(data.status),
            remote_status=ArtifactRemoteStatus(data.remote_status) if data.remote_status else None,
            readme=data.readme,
            version=data.version,
            size=ByteSize(data.size) if data.size is not None else None,
            created_at=data.created_at,
            updated_at=data.updated_at,
            digest=data.digest,
            verification_result=ArtifactVerificationGQLResult.from_dataclass(
                data.verification_result
            )
            if data.verification_result
            else None,
        )

    @strawberry.field
    async def artifact(self, info: Info[StrawberryGQLContext]) -> Artifact:
        revision_action_result = (
            await info.context.processors.artifact_revision.get.wait_for_complete(
                GetArtifactRevisionAction(artifact_revision_id=uuid.UUID(self.id))
            )
        )

        artifact_id = revision_action_result.revision.artifact_id

        artifact_action_result = await info.context.processors.artifact.get.wait_for_complete(
            GetArtifactAction(artifact_id=artifact_id)
        )

        registry_meta_loader = DataLoader(
            apartial(ArtifactRegistryMeta.load_by_id, info.context),
        )

        registry_data = await registry_meta_loader.load(artifact_action_result.result.registry_id)
        source_registry_data = await registry_meta_loader.load(
            artifact_action_result.result.source_registry_id
        )

        return Artifact.from_dataclass(
            artifact_action_result.result, registry_data.url, source_registry_data.url
        )


ArtifactEdge = Edge[Artifact]
ArtifactRevisionEdge = Edge[ArtifactRevision]


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Paginated connection for artifacts with total count information.

    Used for relay-style pagination with cursor-based navigation.
    """)
)
class ArtifactConnection(Connection[Artifact]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Paginated connection for artifact revisions with total count information.

    Used for relay-style pagination with cursor-based navigation.
    """)
)
class ArtifactRevisionConnection(Connection[ArtifactRevision]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Payload for artifact import progress subscription events.

    Provides real-time updates during the artifact import process,
    including progress percentage and current status.
    """)
)
class ArtifactImportProgressUpdatedPayload:
    artifact_id: ID
    progress: float
    status: ArtifactStatus


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact scanning operations.

    Contains the list of artifacts discovered during scanning of external registries.
    These artifacts are registered with SCANNED status and can be imported for actual use.
    """)
)
class ScanArtifactsPayload:
    artifacts: list[Artifact]


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Response payload for delegated artifact scanning operation.
    Contains the list of artifacts discovered during the scan of a reservoir registry's remote registry.
    These artifacts are now available for import or direct use.
""")
)
class DelegateScanArtifactsPayload:
    artifacts: list[Artifact] = strawberry.field(
        description="List of artifacts discovered during the delegated scan from the reservoir registry's remote registry"
    )


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Represents a background task for importing an artifact revision.

    Contains the task ID for monitoring progress and the associated artifact revision
    being imported from external registries.
    """)
)
class ArtifactRevisionImportTask:
    task_id: ID | None
    artifact_revision: ArtifactRevision


# Mutation Payloads
@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact import operations.

    Contains the imported artifact revisions and their associated background tasks.
    Tasks can be monitored to track the import progress from SCANNED to AVAILABLE status.
    """)
)
class ImportArtifactsPayload:
    artifact_revisions: ArtifactRevisionConnection
    tasks: list[ArtifactRevisionImportTask]


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Response payload for delegated artifact import operation.
    Contains the imported artifact revisions and associated background tasks.
    The tasks can be monitored to track the progress of the import operation.
""")
)
class DelegateImportArtifactsPayload:
    artifact_revisions: ArtifactRevisionConnection = strawberry.field(
        description="Connection of artifact revisions that were imported from the reservoir registry's remote registry"
    )
    tasks: list[ArtifactRevisionImportTask] = strawberry.field(
        description="List of background tasks created for importing the artifact revisions"
    )


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact update operations.

    Returns the updated artifact with modified metadata properties.
    """)
)
class UpdateArtifactPayload:
    artifact: Artifact


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact revision cleanup operations.

    Contains the cleaned artifact revisions that have had their stored data removed,
    transitioning them back to SCANNED status to free storage space.
    """)
)
class CleanupArtifactRevisionsPayload:
    artifact_revisions: ArtifactRevisionConnection


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact revision approval operations.

    Contains the approved artifact revision. Admin-only operation.
    """)
)
class ApproveArtifactPayload:
    artifact_revision: ArtifactRevision


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for artifact revision rejection operations.

    Contains the rejected artifact revision. Admin-only operation.
    """)
)
class RejectArtifactPayload:
    artifact_revision: ArtifactRevision


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for canceling artifact import operations.

    Contains the artifact revision whose import was canceled,
    reverting its status back to SCANNED.
    """)
)
class CancelImportArtifactPayload:
    artifact_revision: ArtifactRevision


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Payload for artifact status change subscription events.

    Provides real-time notifications when artifact revision statuses change
    during import, cleanup, or other operations.
    """)
)
class ArtifactStatusChangedPayload:
    artifact_revision: ArtifactRevision


@strawberry.type(
    description=dedent_strip("""
    Added in 25.14.0.

    Response payload for batch model scanning operations.

    Contains the artifact revisions discovered during detailed scanning of specific models,
    including README content and file size information.
    """)
)
class ScanArtifactModelsPayload:
    artifact_revision: ArtifactRevisionConnection


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Response payload for artifact deletion operations.

    Contains the artifacts that were soft-deleted. These can be restored later.
    """)
)
class DeleteArtifactsPayload:
    artifacts: list[Artifact]


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Response payload for artifact restoration operations.

    Contains the artifacts that were restored from soft-deleted state.
    """)
)
class RestoreArtifactsPayload:
    artifacts: list[Artifact]



from .creators import EndpointCreatorSpec
from .repository import ModelServingRepository
from .updaters import EndpointAutoScalingRuleUpdaterSpec, EndpointUpdaterSpec

__all__ = (
    "EndpointAutoScalingRuleUpdaterSpec",
    "EndpointCreatorSpec",
    "EndpointUpdaterSpec",
    "ModelServingRepository",
)



"""Agent related types."""

from dataclasses import dataclass

from ai.backend.common.types import AgentId, ResourceSlot


@dataclass
class AgentMeta:
    """Agent metadata without cached occupancy values."""

    id: AgentId
    addr: str
    architecture: str
    available_slots: ResourceSlot
    scaling_group: str



"""Query conditions and orders for routes."""

from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.data.deployment.types import RouteStatus, RouteTrafficStatus
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class RouteConditions:
    """Query conditions for routes."""

    @staticmethod
    def by_ids(route_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RoutingRow.id.in_(route_ids)

        return inner

    @staticmethod
    def by_endpoint_id(endpoint_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RoutingRow.endpoint == endpoint_id

        return inner

    @staticmethod
    def by_statuses(statuses: list[RouteStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RoutingRow.status.in_(statuses)

        return inner

    @staticmethod
    def by_traffic_statuses(traffic_statuses: list[RouteTrafficStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RoutingRow.traffic_status.in_(traffic_statuses)

        return inner

    @staticmethod
    def by_revision_ids(revision_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RoutingRow.revision.in_(revision_ids)

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RoutingRow.created_at)
                .where(RoutingRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RoutingRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RoutingRow.created_at)
                .where(RoutingRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RoutingRow.created_at > subquery

        return inner


class RouteOrders:
    """Query orders for routes."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RoutingRow.created_at.asc()
        return RoutingRow.created_at.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RoutingRow.status.asc()
        return RoutingRow.status.desc()

    @staticmethod
    def traffic_ratio(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RoutingRow.traffic_ratio.asc()
        return RoutingRow.traffic_ratio.desc()

    @staticmethod
    def id(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RoutingRow.id.asc()
        return RoutingRow.id.desc()



from __future__ import annotations

import uuid
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column

from ai.backend.manager.data.permission.association_scopes_entities import (
    AssociationScopesEntitiesData,
)
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.types import EntityType, RelationType, ScopeType
from ai.backend.manager.models.base import (
    GUID,
    Base,
    StrEnumType,
)


class AssociationScopesEntitiesRow(Base):  # type: ignore[misc]
    __tablename__ = "association_scopes_entities"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("scope_type", "scope_id", "entity_id", name="uq_scope_id_entity_id"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )

    scope_type: Mapped[ScopeType] = mapped_column(
        "scope_type", StrEnumType(ScopeType, length=32), nullable=False
    )
    scope_id: Mapped[str] = mapped_column(
        "scope_id",
        sa.String(64),
        nullable=False,
    )  # e.g., "global", "domain_id", "project_id", "user_id" etc.
    entity_type: Mapped[EntityType] = mapped_column(
        "entity_type", StrEnumType(EntityType, length=32), nullable=False
    )
    entity_id: Mapped[str] = mapped_column(
        "entity_id",
        sa.String(64),
        nullable=False,
    )
    relation_type: Mapped[RelationType] = mapped_column(
        "relation_type",
        StrEnumType(RelationType, length=32),
        nullable=False,
        server_default=RelationType.AUTO.value,
    )
    registered_at: Mapped[datetime] = mapped_column(
        "registered_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )

    def object_id(self) -> ObjectId:
        return ObjectId(entity_type=self.entity_type, entity_id=self.entity_id)

    def parsed_scope_id(self) -> ScopeId:
        """
        Convert the association to a ScopeId.
        """
        return ScopeId(scope_type=self.scope_type, scope_id=self.scope_id)

    def to_data(self) -> AssociationScopesEntitiesData:
        """
        Convert the association to a data object.
        """
        return AssociationScopesEntitiesData(
            id=self.id,
            scope_id=ScopeId(
                scope_type=self.scope_type,
                scope_id=self.scope_id,
            ),
            object_id=self.object_id(),
            relation_type=self.relation_type,
        )



"""add_scheduling_rank_to_user_fair_shares

Revision ID: 352143f82276
Revises: 5bbce2f389e4
Create Date: 2026-01-20 10:29:05.603026

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "352143f82276"
down_revision = "5bbce2f389e4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "user_fair_shares",
        sa.Column(
            "scheduling_rank",
            sa.Integer(),
            nullable=True,
            comment="Computed scheduling priority rank. "
            "Lower value = higher priority (1 = highest). "
            "NULL means rank calculation has not been performed yet.",
        ),
    )


def downgrade() -> None:
    op.drop_column("user_fair_shares", "scheduling_rank")



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.vfs_storage.actions.create import (
    CreateVFSStorageAction,
    CreateVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.delete import (
    DeleteVFSStorageAction,
    DeleteVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.get import (
    GetVFSStorageAction,
    GetVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.list import (
    ListVFSStorageAction,
    ListVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.search import (
    SearchVFSStoragesAction,
    SearchVFSStoragesActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.update import (
    UpdateVFSStorageAction,
    UpdateVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.service import VFSStorageService


class VFSStorageProcessors(AbstractProcessorPackage):
    create: ActionProcessor[CreateVFSStorageAction, CreateVFSStorageActionResult]
    update: ActionProcessor[UpdateVFSStorageAction, UpdateVFSStorageActionResult]
    delete: ActionProcessor[DeleteVFSStorageAction, DeleteVFSStorageActionResult]
    get: ActionProcessor[GetVFSStorageAction, GetVFSStorageActionResult]
    list_storages: ActionProcessor[ListVFSStorageAction, ListVFSStorageActionResult]
    search_vfs_storages: ActionProcessor[SearchVFSStoragesAction, SearchVFSStoragesActionResult]

    def __init__(self, service: VFSStorageService, action_monitors: list[ActionMonitor]) -> None:
        self.create = ActionProcessor(service.create, action_monitors)
        self.update = ActionProcessor(service.update, action_monitors)
        self.delete = ActionProcessor(service.delete, action_monitors)
        self.get = ActionProcessor(service.get, action_monitors)
        self.list_storages = ActionProcessor(service.list, action_monitors)
        self.search_vfs_storages = ActionProcessor(service.search, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateVFSStorageAction.spec(),
            UpdateVFSStorageAction.spec(),
            DeleteVFSStorageAction.spec(),
            GetVFSStorageAction.spec(),
            ListVFSStorageAction.spec(),
            SearchVFSStoragesAction.spec(),
        ]



from __future__ import annotations

import uuid
from functools import cached_property, partial

from strawberry.dataloader import DataLoader

from ai.backend.common.types import AgentId, ImageID, KernelId
from ai.backend.manager.data.agent.types import AgentDetailData
from ai.backend.manager.data.artifact.types import ArtifactData, ArtifactRevisionData
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryData
from ai.backend.manager.data.deployment.types import (
    DeploymentHistoryData,
    ModelDeploymentAccessTokenData,
    ModelDeploymentAutoScalingRuleData,
    ModelDeploymentData,
    ModelReplicaData,
    ModelRevisionData,
    RouteHistoryData,
    RouteInfo,
)
from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.data.image.types import ImageAliasData, ImageData
from ai.backend.manager.data.kernel.types import KernelInfo
from ai.backend.manager.data.notification import NotificationChannelData, NotificationRuleData
from ai.backend.manager.data.object_storage.types import ObjectStorageData
from ai.backend.manager.data.permission.permission import PermissionData
from ai.backend.manager.data.permission.role import AssignedUserData, RoleData
from ai.backend.manager.data.reservoir_registry.types import ReservoirRegistryData
from ai.backend.manager.data.scaling_group.types import ScalingGroupData
from ai.backend.manager.data.session.types import SessionSchedulingHistoryData
from ai.backend.manager.data.storage_namespace.types import StorageNamespaceData
from ai.backend.manager.data.user.types import UserData
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.services.processors import Processors

from .agent import load_agents_by_ids, load_container_counts
from .artifact import load_artifacts_by_ids
from .artifact_registry import load_artifact_registries_by_ids
from .artifact_revision import load_artifact_revisions_by_ids
from .deployment import (
    load_access_tokens_by_ids,
    load_auto_scaling_rules_by_ids,
    load_deployments_by_ids,
    load_replicas_by_ids,
    load_revisions_by_ids,
    load_routes_by_ids,
)
from .domain import load_domains_by_names
from .huggingface_registry import load_huggingface_registries_by_ids
from .image import load_alias_by_ids, load_images_by_ids
from .kernel import load_kernels_by_ids
from .notification import load_channels_by_ids, load_rules_by_ids
from .object_storage import load_object_storages_by_ids
from .project import load_projects_by_ids
from .rbac import (
    load_permissions_by_ids,
    load_role_assignments_by_ids,
    load_role_assignments_by_role_and_user_ids,
    load_roles_by_ids,
)
from .reservoir_registry import load_reservoir_registries_by_ids
from .resource_group import load_resource_groups_by_names
from .scheduling_history import (
    load_deployment_histories_by_ids,
    load_route_histories_by_ids,
    load_session_histories_by_ids,
)
from .storage_namespace import load_storage_namespaces_by_ids
from .user import load_users_by_ids
from .vfs_storage import load_vfs_storages_by_ids


class DataLoaders:
    """
    Manages domain-specific DataLoader instances for GraphQL resolvers.

    This class is the central registry for all DataLoaders used in the GraphQL API.
    Each domain (notification, model_deployment, model_replica, etc.) will have
    its own loader instances initialized here.
    """

    _processors: Processors

    def __init__(self, processors: Processors) -> None:
        self._processors = processors

    @cached_property
    def resource_group_loader(
        self,
    ) -> DataLoader[str, ScalingGroupData | None]:
        return DataLoader(
            load_fn=partial(load_resource_groups_by_names, self._processors.scaling_group)
        )

    @cached_property
    def notification_channel_loader(
        self,
    ) -> DataLoader[uuid.UUID, NotificationChannelData | None]:
        return DataLoader(load_fn=partial(load_channels_by_ids, self._processors.notification))

    @cached_property
    def notification_rule_loader(
        self,
    ) -> DataLoader[uuid.UUID, NotificationRuleData | None]:
        return DataLoader(load_fn=partial(load_rules_by_ids, self._processors.notification))

    @cached_property
    def artifact_registry_loader(
        self,
    ) -> DataLoader[uuid.UUID, ArtifactRegistryData | None]:
        return DataLoader(
            load_fn=partial(load_artifact_registries_by_ids, self._processors.artifact_registry)
        )

    @cached_property
    def huggingface_registry_loader(
        self,
    ) -> DataLoader[uuid.UUID, HuggingFaceRegistryData | None]:
        return DataLoader(
            load_fn=partial(load_huggingface_registries_by_ids, self._processors.artifact_registry)
        )

    @cached_property
    def reservoir_registry_loader(
        self,
    ) -> DataLoader[uuid.UUID, ReservoirRegistryData | None]:
        return DataLoader(
            load_fn=partial(load_reservoir_registries_by_ids, self._processors.artifact_registry)
        )

    @cached_property
    def storage_namespace_loader(
        self,
    ) -> DataLoader[uuid.UUID, StorageNamespaceData | None]:
        return DataLoader(
            load_fn=partial(load_storage_namespaces_by_ids, self._processors.storage_namespace)
        )

    @cached_property
    def object_storage_loader(
        self,
    ) -> DataLoader[uuid.UUID, ObjectStorageData | None]:
        return DataLoader(
            load_fn=partial(load_object_storages_by_ids, self._processors.object_storage)
        )

    @cached_property
    def vfs_storage_loader(
        self,
    ) -> DataLoader[uuid.UUID, VFSStorageData | None]:
        return DataLoader(load_fn=partial(load_vfs_storages_by_ids, self._processors.vfs_storage))

    @cached_property
    def artifact_loader(
        self,
    ) -> DataLoader[uuid.UUID, ArtifactData | None]:
        return DataLoader(load_fn=partial(load_artifacts_by_ids, self._processors.artifact))

    @cached_property
    def artifact_revision_loader(
        self,
    ) -> DataLoader[uuid.UUID, ArtifactRevisionData | None]:
        return DataLoader(
            load_fn=partial(load_artifact_revisions_by_ids, self._processors.artifact_revision)
        )

    @cached_property
    def route_loader(
        self,
    ) -> DataLoader[uuid.UUID, RouteInfo | None]:
        return DataLoader(load_fn=partial(load_routes_by_ids, self._processors.deployment))

    @cached_property
    def deployment_loader(
        self,
    ) -> DataLoader[uuid.UUID, ModelDeploymentData | None]:
        return DataLoader(load_fn=partial(load_deployments_by_ids, self._processors.deployment))

    @cached_property
    def revision_loader(
        self,
    ) -> DataLoader[uuid.UUID, ModelRevisionData | None]:
        return DataLoader(load_fn=partial(load_revisions_by_ids, self._processors.deployment))

    @cached_property
    def replica_loader(
        self,
    ) -> DataLoader[uuid.UUID, ModelReplicaData | None]:
        return DataLoader(load_fn=partial(load_replicas_by_ids, self._processors.deployment))

    @cached_property
    def container_count_loader(
        self,
    ) -> DataLoader[AgentId, int]:
        return DataLoader(load_fn=partial(load_container_counts, self._processors.agent))

    @cached_property
    def image_loader(
        self,
    ) -> DataLoader[ImageID, ImageData | None]:
        return DataLoader(load_fn=partial(load_images_by_ids, self._processors.image))

    @cached_property
    def kernel_loader(
        self,
    ) -> DataLoader[KernelId, KernelInfo | None]:
        return DataLoader(load_fn=partial(load_kernels_by_ids, self._processors.session))

    @cached_property
    def image_alias_loader(
        self,
    ) -> DataLoader[uuid.UUID, ImageAliasData | None]:
        """Load a single alias by its own ID (ImageAliasRow.id)."""
        return DataLoader(load_fn=partial(load_alias_by_ids, self._processors.image))

    @cached_property
    def user_loader(
        self,
    ) -> DataLoader[uuid.UUID, UserData | None]:
        return DataLoader(load_fn=partial(load_users_by_ids, self._processors.user))

    @cached_property
    def domain_loader(
        self,
    ) -> DataLoader[str, DomainData | None]:
        return DataLoader(load_fn=partial(load_domains_by_names, self._processors.domain))

    @cached_property
    def project_loader(
        self,
    ) -> DataLoader[uuid.UUID, GroupData | None]:
        return DataLoader(load_fn=partial(load_projects_by_ids, self._processors.group))

    @cached_property
    def agent_loader(
        self,
    ) -> DataLoader[AgentId, AgentDetailData | None]:
        return DataLoader(load_fn=partial(load_agents_by_ids, self._processors.agent))

    @cached_property
    def auto_scaling_rule_loader(
        self,
    ) -> DataLoader[uuid.UUID, ModelDeploymentAutoScalingRuleData | None]:
        return DataLoader(
            load_fn=partial(load_auto_scaling_rules_by_ids, self._processors.deployment)
        )

    @cached_property
    def access_token_loader(
        self,
    ) -> DataLoader[uuid.UUID, ModelDeploymentAccessTokenData | None]:
        return DataLoader(load_fn=partial(load_access_tokens_by_ids, self._processors.deployment))

    @cached_property
    def session_history_loader(
        self,
    ) -> DataLoader[uuid.UUID, SessionSchedulingHistoryData | None]:
        return DataLoader(
            load_fn=partial(load_session_histories_by_ids, self._processors.scheduling_history)
        )

    @cached_property
    def deployment_history_loader(
        self,
    ) -> DataLoader[uuid.UUID, DeploymentHistoryData | None]:
        return DataLoader(
            load_fn=partial(load_deployment_histories_by_ids, self._processors.scheduling_history)
        )

    @cached_property
    def route_history_loader(
        self,
    ) -> DataLoader[uuid.UUID, RouteHistoryData | None]:
        return DataLoader(
            load_fn=partial(load_route_histories_by_ids, self._processors.scheduling_history)
        )

    @cached_property
    def role_loader(
        self,
    ) -> DataLoader[uuid.UUID, RoleData | None]:
        return DataLoader(
            load_fn=partial(load_roles_by_ids, self._processors.permission_controller)
        )

    @cached_property
    def permission_loader(
        self,
    ) -> DataLoader[uuid.UUID, PermissionData | None]:
        return DataLoader(
            load_fn=partial(load_permissions_by_ids, self._processors.permission_controller)
        )

    @cached_property
    def role_assignment_loader(
        self,
    ) -> DataLoader[uuid.UUID, AssignedUserData | None]:
        return DataLoader(
            load_fn=partial(load_role_assignments_by_ids, self._processors.permission_controller)
        )

    @cached_property
    def role_assignment_by_role_and_user_loader(
        self,
    ) -> DataLoader[tuple[uuid.UUID, uuid.UUID], AssignedUserData | None]:
        return DataLoader(
            load_fn=partial(
                load_role_assignments_by_role_and_user_ids,
                self._processors.permission_controller,
            )
        )



"""Database source for container registry quota repository operations."""

from __future__ import annotations

from typing import Any
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import load_only

from ai.backend.manager.data.container_registry.types import PerProjectContainerRegistryInfo
from ai.backend.manager.errors.image import ContainerRegistryNotFound
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.rbac import ProjectScope
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


class PerProjectRegistryQuotaDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def fetch_container_registry_row(
        self, scope_id: ProjectScope
    ) -> PerProjectContainerRegistryInfo:
        async with self._db.begin_readonly_session() as db_sess:
            project_id = scope_id.project_id
            project_row = await self._fetch_project_row(db_sess, project_id)

            if project_row is None:
                raise ContainerRegistryNotFound(
                    f"Container registry info does not exist or is invalid in the project. (project: {project_id})"
                )
            container_registry: dict[str, Any] | None = project_row.container_registry
            if (
                not container_registry
                or "registry" not in container_registry
                or "project" not in container_registry
            ):
                raise ContainerRegistryNotFound(
                    f"Container registry info does not exist or is invalid in the project. (project: {project_id})"
                )
            registry_name, project = (
                container_registry["registry"],
                container_registry["project"],
            )

            registry_row = await self._fetch_registry_row(db_sess, registry_name, project)

            if registry_row is None:
                raise ContainerRegistryNotFound(
                    f"Container registry row not found. (registry: {registry_name}, project: {project})"
                )

            return PerProjectContainerRegistryInfo(
                id=registry_row.id,
                url=registry_row.url,
                registry_name=registry_row.registry_name,
                type=registry_row.type,
                project=registry_row.project or "",
                username=registry_row.username or "",
                password=registry_row.password or "",
                ssl_verify=registry_row.ssl_verify if registry_row.ssl_verify is not None else True,
                is_global=registry_row.is_global if registry_row.is_global is not None else False,
                extra=registry_row.extra or {},
            )

    async def _fetch_project_row(
        self,
        db_sess: SASession,
        project_id: UUID,
    ) -> GroupRow | None:
        project_query = (
            sa.select(GroupRow)
            .where(GroupRow.id == project_id)
            .options(load_only(GroupRow.container_registry))
        )
        result = await db_sess.execute(project_query)
        return result.scalar_one_or_none()

    async def _fetch_registry_row(
        self,
        db_sess: SASession,
        registry_name: str,
        project: str,
    ) -> ContainerRegistryRow | None:
        registry_query = sa.select(ContainerRegistryRow).where(
            (ContainerRegistryRow.registry_name == registry_name)
            & (ContainerRegistryRow.project == project)
        )
        result = await db_sess.execute(registry_query)
        return result.scalars().one_or_none()



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

import sqlalchemy as sa

from ai.backend.common.types import AccessKey
from ai.backend.manager.models.scaling_group import (
    ScalingGroupForDomainRow,
    ScalingGroupForKeypairsRow,
    ScalingGroupForProjectRow,
)
from ai.backend.manager.repositories.base.purger import BatchPurger, BatchPurgerSpec


@dataclass
class ScalingGroupForDomainPurgerSpec(BatchPurgerSpec[ScalingGroupForDomainRow]):
    """PurgerSpec for disassociating a scaling group from a domain."""

    scaling_group: str
    domain: str

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForDomainRow]]:
        return sa.select(ScalingGroupForDomainRow).where(
            sa.and_(
                ScalingGroupForDomainRow.scaling_group == self.scaling_group,
                ScalingGroupForDomainRow.domain == self.domain,
            )
        )


@dataclass
class ScalingGroupsForDomainPurgerSpec(BatchPurgerSpec[ScalingGroupForDomainRow]):
    """PurgerSpec for disassociating multiple scaling groups from a domain."""

    scaling_groups: list[str]
    domain: str

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForDomainRow]]:
        return sa.select(ScalingGroupForDomainRow).where(
            sa.and_(
                ScalingGroupForDomainRow.scaling_group.in_(self.scaling_groups),
                ScalingGroupForDomainRow.domain == self.domain,
            )
        )


@dataclass
class AllScalingGroupsForDomainPurgerSpec(BatchPurgerSpec[ScalingGroupForDomainRow]):
    """PurgerSpec for disassociating all scaling groups from a domain."""

    domain: str

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForDomainRow]]:
        return sa.select(ScalingGroupForDomainRow).where(
            ScalingGroupForDomainRow.domain == self.domain,
        )


@dataclass
class ScalingGroupForKeypairsPurgerSpec(BatchPurgerSpec[ScalingGroupForKeypairsRow]):
    """PurgerSpec for disassociating a scaling group from a keypair."""

    scaling_group: str
    access_key: AccessKey

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForKeypairsRow]]:
        return sa.select(ScalingGroupForKeypairsRow).where(
            sa.and_(
                ScalingGroupForKeypairsRow.scaling_group == self.scaling_group,
                ScalingGroupForKeypairsRow.access_key == self.access_key,
            )
        )


@dataclass
class ScalingGroupsForKeypairsPurgerSpec(BatchPurgerSpec[ScalingGroupForKeypairsRow]):
    """PurgerSpec for disassociating multiple scaling groups from a keypair."""

    scaling_groups: list[str]
    access_key: AccessKey

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForKeypairsRow]]:
        return sa.select(ScalingGroupForKeypairsRow).where(
            sa.and_(
                ScalingGroupForKeypairsRow.scaling_group.in_(self.scaling_groups),
                ScalingGroupForKeypairsRow.access_key == self.access_key,
            )
        )


def create_scaling_group_for_domain_purger(
    scaling_group: str,
    domain: str,
) -> BatchPurger[ScalingGroupForDomainRow]:
    """Create a BatchPurger for disassociating a scaling group from a domain."""
    return BatchPurger(
        spec=ScalingGroupForDomainPurgerSpec(
            scaling_group=scaling_group,
            domain=domain,
        ),
        batch_size=1,  # We expect only one row to be deleted
    )


def create_scaling_group_for_keypairs_purger(
    scaling_group: str,
    access_key: AccessKey,
) -> BatchPurger[ScalingGroupForKeypairsRow]:
    """Create a BatchPurger for disassociating a scaling group from a keypair."""
    return BatchPurger(
        spec=ScalingGroupForKeypairsPurgerSpec(
            scaling_group=scaling_group,
            access_key=access_key,
        ),
        batch_size=1,  # We expect only one row to be deleted
    )


@dataclass
class ScalingGroupForProjectPurgerSpec(BatchPurgerSpec[ScalingGroupForProjectRow]):
    """PurgerSpec for disassociating a scaling group from a project (user group)."""

    scaling_group: str
    project: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForProjectRow]]:
        return sa.select(ScalingGroupForProjectRow).where(
            sa.and_(
                ScalingGroupForProjectRow.scaling_group == self.scaling_group,
                ScalingGroupForProjectRow.group == self.project,
            )
        )


@dataclass
class ScalingGroupsForProjectPurgerSpec(BatchPurgerSpec[ScalingGroupForProjectRow]):
    """PurgerSpec for disassociating multiple scaling groups from a project (user group)."""

    scaling_groups: list[str]
    project: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForProjectRow]]:
        return sa.select(ScalingGroupForProjectRow).where(
            sa.and_(
                ScalingGroupForProjectRow.scaling_group.in_(self.scaling_groups),
                ScalingGroupForProjectRow.group == self.project,
            )
        )


@dataclass
class AllScalingGroupsForProjectPurgerSpec(BatchPurgerSpec[ScalingGroupForProjectRow]):
    """PurgerSpec for disassociating all scaling groups from a project (user group)."""

    project: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ScalingGroupForProjectRow]]:
        return sa.select(ScalingGroupForProjectRow).where(
            ScalingGroupForProjectRow.group == self.project,
        )


def create_scaling_group_for_project_purger(
    scaling_group: str,
    project: UUID,
) -> BatchPurger[ScalingGroupForProjectRow]:
    """Create a BatchPurger for disassociating a scaling group from a project."""
    return BatchPurger(
        spec=ScalingGroupForProjectPurgerSpec(
            scaling_group=scaling_group,
            project=project,
        ),
        batch_size=1,  # We expect only one row to be deleted
    )



from __future__ import annotations

import logging
import uuid
from collections.abc import Callable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Self
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship
from sqlalchemy.sql.dml import Delete, Update
from sqlalchemy.sql.selectable import Select

from ai.backend.common.types import ResourceSlot
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.models.base import (
    GUID,
    Base,
    ResourceSlotColumn,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.scaling_group import ScalingGroupRow

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models"))

__all__: Sequence[str] = ("resource_presets",)


# Type alias for statements that support .where() method
type WhereableStatement[T] = Select[T] | Update | Delete


def filter_by_name(name: str) -> Callable[[WhereableStatement[Any]], WhereableStatement[Any]]:
    return lambda query_stmt: query_stmt.where(ResourcePresetRow.name == name)


def filter_by_id(id: UUID) -> Callable[[WhereableStatement[Any]], WhereableStatement[Any]]:
    return lambda query_stmt: query_stmt.where(ResourcePresetRow.id == id)


# QueryOption is a function that takes a statement and returns a filtered statement
type QueryOption = Callable[[WhereableStatement[Any]], WhereableStatement[Any]]


def _get_scaling_group_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.scaling_group import ScalingGroupRow

    return ScalingGroupRow.name == foreign(ResourcePresetRow.scaling_group_name)


class ResourcePresetRow(Base):  # type: ignore[misc]
    __tablename__ = "resource_presets"
    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    name: Mapped[str] = mapped_column("name", sa.String(length=256), nullable=False)
    resource_slots: Mapped[ResourceSlot] = mapped_column(
        "resource_slots", ResourceSlotColumn(), nullable=False
    )
    shared_memory: Mapped[int | None] = mapped_column(
        "shared_memory", sa.BigInteger(), nullable=True
    )

    # If `scaling_group_name` is None, the preset is global
    scaling_group_name: Mapped[str | None] = mapped_column(
        "scaling_group_name", sa.String(length=64), nullable=True, server_default=sa.null()
    )
    scaling_group_row: Mapped[ScalingGroupRow | None] = relationship(
        "ScalingGroupRow",
        back_populates="resource_preset_rows",
        primaryjoin=_get_scaling_group_join_condition,
    )

    __table_args__ = (
        sa.Index(
            "ix_resource_presets_name_null_scaling_group_name",
            name,
            postgresql_where=scaling_group_name.is_(None),
            unique=True,
        ),
        sa.Index(
            "ix_resource_presets_name_scaling_group_name",
            name,
            scaling_group_name,
            postgresql_where=scaling_group_name.isnot(None),
            unique=True,
        ),
    )

    @classmethod
    async def update(
        cls,
        query_option: QueryOption,
        data: Mapping[str, Any],
        *,
        db_session: AsyncSession,
    ) -> Self | None:
        base_update_stmt = sa.update(ResourcePresetRow).values(data).returning(ResourcePresetRow)
        filtered_stmt = query_option(base_update_stmt)
        stmt = (
            sa.select(ResourcePresetRow)
            .from_statement(filtered_stmt)
            .execution_options(populate_existing=True)
        )
        try:
            result: Self | None = await db_session.scalar(stmt)
            return result
        except sa.exc.IntegrityError:
            return None

    @classmethod
    async def delete(
        cls,
        query_option: QueryOption,
        *,
        db_session: AsyncSession,
    ) -> None:
        base_delete_stmt = sa.delete(ResourcePresetRow)
        filtered_stmt = query_option(base_delete_stmt)
        await db_session.execute(filtered_stmt)

    def to_dataclass(self) -> ResourcePresetData:
        return ResourcePresetData(
            id=self.id,
            name=self.name,
            resource_slots=self.resource_slots,
            shared_memory=self.shared_memory,
            scaling_group_name=self.scaling_group_name,
        )


# For compatibility
resource_presets = ResourcePresetRow.__table__



from dataclasses import dataclass
from typing import override

from ai.backend.common.types import ImageID
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.image.types import ImageData, ResourceLimitInput
from ai.backend.manager.services.image.actions.resource_limit_base import ImageResourceLimitAction


@dataclass
class SetImageResourceLimitByIdAction(ImageResourceLimitAction):
    image_id: ImageID
    resource_limit: ResourceLimitInput

    @override
    def entity_id(self) -> str | None:
        return str(self.image_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class SetImageResourceLimitByIdActionResult(BaseActionResult):
    image_data: ImageData

    @override
    def entity_id(self) -> str | None:
        return str(self.image_data.id)



"""Creator for RBAC field-scoped entity insert operations."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import TypeVar

from sqlalchemy import inspect
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.data.permission.types import EntityType, FieldType
from ai.backend.manager.errors.repository import UnsupportedCompositePrimaryKeyError
from ai.backend.manager.models.base import Base
from ai.backend.manager.models.rbac_models.entity_field import EntityFieldRow
from ai.backend.manager.repositories.base.creator import CreatorSpec

TRow = TypeVar("TRow", bound=Base)


# =============================================================================
# Single Field Creator
# =============================================================================


@dataclass
class RBACFieldCreator[TRow: Base]:
    """Creator for a single field-scoped entity.

    Attributes:
        spec: CreatorSpec implementation defining the row to create.
        entity_type: The entity type of the parent entity.
        entity_id: The ID of the parent entity.
        field_type: The field type for RBAC field mapping.
    """

    spec: CreatorSpec[TRow]
    entity_type: EntityType
    entity_id: str
    field_type: FieldType


@dataclass
class RBACFieldCreatorResult[TRow: Base]:
    """Result of executing a single field creation."""

    row: TRow


async def execute_rbac_field_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACFieldCreator[TRow],
) -> RBACFieldCreatorResult[TRow]:
    """Create a field-scoped entity with its entity-field mapping.

    Operations:
    1. Insert main field row
    2. Flush to get DB-generated ID
    3. Extract field ID from row
    4. Insert EntityFieldRow (parent_entity -> field mapping)

    The EntityFieldRow maps the field to its parent entity,
    enabling entity-field relationship tracking.

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Creator instance with spec defining the field to create.

    Returns:
        RBACFieldCreatorResult containing the created row.
    """
    spec = creator.spec
    row = spec.build_row()
    mapper = inspect(type(row))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Creator only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 1. Build and insert row
    db_sess.add(row)

    # 2. Flush to get DB-generated ID
    await db_sess.flush()

    # 3. Extract field ID and insert entity-field mapping
    instance_state = inspect(row)
    pk_value = instance_state.identity[0]
    db_sess.add(
        EntityFieldRow(
            entity_type=creator.entity_type,
            entity_id=creator.entity_id,
            field_type=creator.field_type,
            field_id=str(pk_value),
        )
    )

    return RBACFieldCreatorResult(row=row)


# =============================================================================
# Bulk Field Creator
# =============================================================================


@dataclass
class RBACBulkFieldCreator[TRow: Base]:
    """Bulk creator for multiple field-scoped entities.

    Attributes:
        specs: Sequence of CreatorSpec implementations.
        entity_type: The entity type of the parent entity for all fields.
        entity_id: The ID of the parent entity for all fields.
        field_type: The field type for all fields.
    """

    specs: Sequence[CreatorSpec[TRow]]
    entity_type: EntityType
    entity_id: str
    field_type: FieldType


@dataclass
class RBACBulkFieldCreatorResult[TRow: Base]:
    """Result of executing a bulk field creation."""

    rows: list[TRow]


async def execute_rbac_bulk_field_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACBulkFieldCreator[TRow],
) -> RBACBulkFieldCreatorResult[TRow]:
    """Create multiple field-scoped entities in a single transaction.

    Operations:
    1. Build and insert all field rows
    2. Flush to get DB-generated IDs
    3. Bulk insert EntityFieldRows

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Bulk creator with specs defining fields to create.

    Returns:
        RBACBulkFieldCreatorResult containing all created rows.
    """
    if not creator.specs:
        return RBACBulkFieldCreatorResult(rows=[])

    # 1. Build and add all rows
    rows = [spec.build_row() for spec in creator.specs]
    db_sess.add_all(rows)

    mapper = inspect(type(rows[0]))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Creator only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 2. Flush to get DB-generated IDs and insert field mappings
    await db_sess.flush()

    field_rows = [
        EntityFieldRow(
            entity_type=creator.entity_type,
            entity_id=creator.entity_id,
            field_type=creator.field_type,
            field_id=str(inspect(row).identity[0]),
        )
        for row in rows
    ]
    db_sess.add_all(field_rows)

    return RBACBulkFieldCreatorResult(rows=rows)



"""
ImageV2 GQL fetcher functions for Strawberry GraphQL.

This module provides data fetching logic for ImageV2 queries.
"""

from __future__ import annotations

import uuid
from functools import lru_cache

import strawberry
from strawberry import Info

from ai.backend.common.types import ImageID
from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.image.row import ImageAliasRow, ImageRow
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.services.image.actions.search_aliases import SearchAliasesAction
from ai.backend.manager.services.image.actions.search_images import SearchImagesAction

from .types import (
    ImageV2AliasConnectionGQL,
    ImageV2AliasEdgeGQL,
    ImageV2AliasFilterGQL,
    ImageV2AliasGQL,
    ImageV2AliasOrderByGQL,
    ImageV2ConnectionGQL,
    ImageV2EdgeGQL,
    ImageV2FilterGQL,
    ImageV2GQL,
    ImageV2OrderByGQL,
)


@lru_cache(maxsize=1)
def _get_image_pagination_spec() -> PaginationSpec:
    """Get pagination spec for Image queries."""
    return PaginationSpec(
        forward_order=ImageRow.id.asc(),
        backward_order=ImageRow.id.desc(),
        forward_condition_factory=lambda cursor_value: lambda: ImageRow.id > cursor_value,
        backward_condition_factory=lambda cursor_value: lambda: ImageRow.id < cursor_value,
        tiebreaker_order=ImageRow.id.asc(),
    )


async def fetch_images(
    info: Info[StrawberryGQLContext],
    filter: ImageV2FilterGQL | None = None,
    order_by: list[ImageV2OrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> ImageV2ConnectionGQL:
    """Fetch images with optional filtering, ordering, and pagination.

    Args:
        info: GraphQL context info
        filter: Optional filter criteria
        order_by: Optional ordering specification
        before/after/first/last: Cursor-based pagination parameters
        limit/offset: Offset-based pagination parameters
        base_conditions: Additional conditions to prepend
    """
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        _get_image_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = await info.context.processors.image.search_images.wait_for_complete(
        SearchImagesAction(querier=querier)
    )

    edges = []
    for image_data in action_result.data:
        image = ImageV2GQL.from_data(image_data)
        cursor = encode_cursor(image_data.id)
        edges.append(ImageV2EdgeGQL(node=image, cursor=cursor))

    page_info = strawberry.relay.PageInfo(
        has_next_page=action_result.has_next_page,
        has_previous_page=action_result.has_previous_page,
        start_cursor=edges[0].cursor if edges else None,
        end_cursor=edges[-1].cursor if edges else None,
    )

    return ImageV2ConnectionGQL(
        count=action_result.total_count,
        edges=edges,
        page_info=page_info,
    )


async def fetch_image(
    info: Info[StrawberryGQLContext],
    image_id: ImageID,
) -> ImageV2GQL | None:
    """Fetch a single image by ID using dataloader.

    Args:
        info: GraphQL context info
        image_id: The ImageID of the image to fetch

    Returns:
        ImageV2GQL if found, None otherwise
    """
    image_data = await info.context.data_loaders.image_loader.load(image_id)
    if image_data is None:
        return None
    return ImageV2GQL.from_data(image_data)


# =============================================================================
# Image Alias Fetchers
# =============================================================================


async def fetch_image_alias(
    info: Info[StrawberryGQLContext],
    alias_id: uuid.UUID,
) -> ImageV2AliasGQL | None:
    """Fetch a single image alias by ID using dataloader.

    Args:
        info: GraphQL context info
        alias_id: The UUID of the alias to fetch

    Returns:
        ImageV2AliasGQL if found, None otherwise
    """
    alias_data = await info.context.data_loaders.image_alias_loader.load(alias_id)
    if alias_data is None:
        return None
    return ImageV2AliasGQL.from_data(alias_data)


@lru_cache(maxsize=1)
def _get_image_alias_pagination_spec() -> PaginationSpec:
    """Get pagination spec for ImageAlias queries."""
    return PaginationSpec(
        forward_order=ImageAliasRow.id.asc(),
        backward_order=ImageAliasRow.id.desc(),
        forward_condition_factory=lambda cursor_value: lambda: ImageAliasRow.id > cursor_value,
        backward_condition_factory=lambda cursor_value: lambda: ImageAliasRow.id < cursor_value,
        tiebreaker_order=ImageAliasRow.id.asc(),
    )


async def fetch_image_aliases(
    info: Info[StrawberryGQLContext],
    filter: ImageV2AliasFilterGQL | None = None,
    order_by: list[ImageV2AliasOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> ImageV2AliasConnectionGQL:
    """Fetch image aliases with optional filtering, ordering, and pagination.

    Args:
        info: GraphQL context info
        filter: Optional filter criteria
        order_by: Optional ordering specification
        before/after/first/last: Cursor-based pagination parameters
        limit/offset: Offset-based pagination parameters
        base_conditions: Additional conditions to prepend
    """
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        _get_image_alias_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = await info.context.processors.image.search_aliases.wait_for_complete(
        SearchAliasesAction(querier=querier)
    )

    edges = []
    for alias_data in action_result.data:
        alias_node = ImageV2AliasGQL.from_data(alias_data)
        cursor = encode_cursor(alias_data.id)
        edges.append(ImageV2AliasEdgeGQL(node=alias_node, cursor=cursor))

    page_info = strawberry.relay.PageInfo(
        has_next_page=action_result.has_next_page,
        has_previous_page=action_result.has_previous_page,
        start_cursor=edges[0].cursor if edges else None,
        end_cursor=edges[-1].cursor if edges else None,
    )

    return ImageV2AliasConnectionGQL(
        count=action_result.total_count,
        edges=edges,
        page_info=page_info,
    )



"""Workload data types for scheduling."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from uuid import UUID

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    ClusterMode,
    ResourceSlot,
    SessionId,
    SessionResult,
    SessionTypes,
)
from ai.backend.manager.models.session import SessionStatus


@dataclass(frozen=True)
class KeyPairResourcePolicy:
    """Resource policy for a keypair."""

    name: str
    total_resource_slots: ResourceSlot
    max_concurrent_sessions: int | None
    max_concurrent_sftp_sessions: int | None
    max_pending_session_count: int | None
    max_pending_session_resource_slots: ResourceSlot | None


@dataclass(frozen=True)
class UserResourcePolicy:
    """Resource policy for a user."""

    name: str
    total_resource_slots: ResourceSlot


@dataclass(frozen=True)
class PendingSessionInfo:
    """Information about a pending session."""

    session_id: SessionId
    requested_slots: ResourceSlot
    creation_time: datetime


@dataclass(frozen=True)
class SessionDependencyInfo:
    """Information about a session dependency."""

    depends_on: SessionId
    dependency_name: str
    dependency_status: SessionStatus
    dependency_result: SessionResult


@dataclass(frozen=True)
class KernelWorkload:
    """Represents a kernel workload within a session."""

    # Unique identifier of the kernel
    kernel_id: UUID
    # Image name for the kernel
    image: str
    # Architecture required for the kernel
    architecture: str
    # Resource requirements for this kernel
    requested_slots: ResourceSlot


@dataclass(frozen=True)
class SessionWorkload:
    """Represents a session workload for scheduling with minimal required fields."""

    # Session identifier
    session_id: SessionId
    # User identification for fairness calculation
    access_key: AccessKey
    # Resource requirements
    requested_slots: ResourceSlot
    # User UUID for user resource limit checks
    user_uuid: UUID
    # Group ID for group resource limit checks
    group_id: UUID
    # Domain name for domain resource limit checks
    domain_name: str
    # Scaling group name
    scaling_group: str
    # Priority level (higher value = higher priority)
    priority: int = 0
    # Session type (INTERACTIVE, BATCH, INFERENCE)
    session_type: SessionTypes = SessionTypes.INTERACTIVE
    # Cluster mode (SINGLE_NODE or MULTI_NODE)
    cluster_mode: ClusterMode = ClusterMode.SINGLE_NODE
    # Scheduled start time for batch sessions
    starts_at: datetime | None = None
    # Whether this is a private session (SFTP)
    is_private: bool = False
    # Kernels to be scheduled for this session
    kernels: list[KernelWorkload] = field(default_factory=list)
    # Manually designated agent (for superadmin)
    designated_agent_ids: list[AgentId] | None = None
    # Kernel counts at endpoint for each agent (for inference session spreading)
    # Only populated for inference sessions with enforce_spreading_endpoint_replica
    kernel_counts_at_endpoint: dict[AgentId, int] | None = None



from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.models.audit_log import AuditLogRow
from ai.backend.manager.repositories.base import QueryCondition


class AuditLogConditions:
    """Query conditions for audit logs."""

    @staticmethod
    def by_ids(audit_log_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AuditLogRow.id.in_(audit_log_ids)

        return inner



"""Audit log repository module."""

from .creators import AuditLogCreatorSpec
from .options import AuditLogConditions
from .repositories import AuditLogRepositories
from .repository import AuditLogRepository

__all__ = (
    "AuditLogConditions",
    "AuditLogCreatorSpec",
    "AuditLogRepositories",
    "AuditLogRepository",
)



import logging
import uuid
from collections.abc import Awaitable, Callable, Sequence
from datetime import UTC, datetime

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.actions.action import (
    BaseActionTriggerMeta,
)
from ai.backend.manager.actions.action.batch import (
    BaseBatchAction,
    BaseBatchActionResult,
)
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.validator.batch import BatchActionValidator

from .base import ActionRunner

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class BatchActionProcessor[
    TBatchAction: BaseBatchAction,
    TBatchActionResult: BaseBatchActionResult,
]:
    _validators: Sequence[BatchActionValidator]

    _runner: ActionRunner[TBatchAction, TBatchActionResult]

    def __init__(
        self,
        func: Callable[[TBatchAction], Awaitable[TBatchActionResult]],
        monitors: Sequence[ActionMonitor] | None = None,
        validators: Sequence[BatchActionValidator] | None = None,
    ) -> None:
        self._runner = ActionRunner(func, monitors)

        self._validators = validators or []

    async def _run(self, action: TBatchAction) -> TBatchActionResult:
        started_at = datetime.now(UTC)
        action_id = uuid.uuid4()
        action_trigger_meta = BaseActionTriggerMeta(action_id=action_id, started_at=started_at)
        for validator in self._validators:
            await validator.validate(action, action_trigger_meta)

        return await self._runner.run(action, action_trigger_meta)

    async def wait_for_complete(self, action: TBatchAction) -> TBatchActionResult:
        return await self._run(action)



from typing import Any
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo

from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql_legacy.gql_relay import AsyncNode
from ai.backend.manager.data.deployment.types import ExtraVFolderMountData


@strawberry.federation.type(keys=["id"], name="VirtualFolderNode", extend=True)
class VFolder:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID, info: Info) -> "VFolder":
        return cls(id=id)


@strawberry.type
class ExtraVFolderMount(Node):
    id: NodeID[str]
    mount_destination: str
    _vfolder_id: strawberry.Private[UUID]

    @strawberry.field
    async def vfolder(self, info: Info[StrawberryGQLContext]) -> VFolder:
        vfolder_global_id = AsyncNode.to_global_id("VirtualFolderNode", str(self._vfolder_id))
        return VFolder(id=ID(vfolder_global_id))

    @classmethod
    def from_dataclass(cls, data: ExtraVFolderMountData) -> "ExtraVFolderMount":
        return cls(
            # TODO: fix id generation logic
            id=ID(f"{data.vfolder_id}:{data.mount_destination}"),
            mount_destination=data.mount_destination,
            _vfolder_id=data.vfolder_id,
        )


ExtraVFolderMountEdge = Edge[ExtraVFolderMount]


@strawberry.type(description="Added in 25.16.0")
class ExtraVFolderMountConnection(Connection[ExtraVFolderMount]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count

    @classmethod
    def from_dataclass(
        cls, mounts_data: list[ExtraVFolderMountData]
    ) -> "ExtraVFolderMountConnection":
        nodes = [ExtraVFolderMount.from_dataclass(data) for data in mounts_data]
        edges = [Edge(node=node, cursor=str(node.id)) for node in nodes]
        page_info = PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        )

        return cls(count=len(nodes), edges=edges, page_info=page_info)
